import io
import sys
from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager
from shutil import copyfileobj
from typing import (
    Optional,
    Tuple,
    Sequence,
    IO,
    Union,
    Any,
    Callable,
    Generator,
)

import subprocess  # nosec

# Here bandit warns about:
# B404:blacklist import_subprocess
# B603:subprocess_without_shell_equals_true
# 1. We do not use shell=True in this package
# 2. All calls to subprocess start with gpg, so possibly unsanitized input would
#    only result in a gpg error message

ISource = Union[bytes, IO[bytes], Callable[[IO[bytes]], None]]
OSource = Union[int, IO[bytes], Callable[[IO[bytes]], None]]
# TODO: unquote the "subprocess.Popen[bytes]" type hint once we enforce python 3.9 (multiple occurrences)

if sys.platform == "win32":

    def startupinfo():
        _startupinfo = subprocess.STARTUPINFO()
        _startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        _startupinfo.wShowWindow = subprocess.SW_HIDE
        return _startupinfo


else:

    def startupinfo() -> bool:  # Literal[None] not supported by mypy
        return False


class GPGError(Exception):
    """Error class that forwards errors returned by calls to GnuPG."""


def raise_if_not_none(e: Optional[BaseException]) -> None:
    if e is not None:
        raise e


def cmd_devnull(
    command: Tuple[str, ...],
    src: Optional[ISource] = None,
    stdout: OSource = subprocess.DEVNULL,
    **kwargs: Any,
) -> None:
    with cmd_pipe(command, src=src, stdout=stdout, **kwargs):
        pass


@contextmanager
def cmd_pipe_stdout(*args: Any, **kwargs: Any) -> Generator[IO[bytes], None, None]:
    with cmd_pipe(*args, **kwargs) as proc:
        yield assert_io(proc.stdout)


@contextmanager
def cmd_pipe(
    command: Tuple[str, ...],
    src: Optional[ISource] = None,
    stdout: OSource = subprocess.PIPE,
    passphrase: Optional[str] = None,
    **kwargs: Any,
) -> Generator["subprocess.Popen[bytes]", None, None]:
    feed = None
    if src is not None:
        if has_fileno(src) and passphrase is None:
            kwargs["stdin"] = src
        else:
            kwargs["stdin"] = subprocess.PIPE
            feed = feed_from_any(src)
    sink = None
    if callable(stdout):
        sink = stdout
        stdout = subprocess.PIPE

    # For the bandit vulnerability suppressed here (nosec):
    # see info close the subprocess import
    with subprocess.Popen(  # nosec
        command,
        stderr=subprocess.PIPE,
        stdout=stdout,
        startupinfo=startupinfo() or None,
        **kwargs,
    ) as proc:
        if proc.stdin is not None:
            if passphrase is not None:
                proc.stdin.write(passphrase.encode() + b"\n")
                proc.stdin.flush()
            if feed is not None and sink is not None:
                with ThreadPoolExecutor(max_workers=2) as executor:
                    future_feed = executor.submit(feed, proc.stdin)
                    future_sink = executor.submit(sink, proc.stdout)
                    future_feed.add_done_callback(
                        lambda _: proc.stdin and proc.stdin.close()
                    )
                    raise_if_not_none(future_feed.exception())
                    raise_if_not_none(future_sink.exception())
            elif feed is not None:
                feed(proc.stdin)
                proc.stdin.close()
            elif sink is not None and proc.stdout is not None:
                sink(proc.stdout)
        yield proc
        err = proc.stderr.read() if proc.stderr is not None else b""
    if proc.returncode != 0:
        raise GPGError(err.decode("utf-8", "replace"))


def stderr_lookahead(proc: "subprocess.Popen[bytes]") -> bytes:
    if proc.stderr is not None:
        out: bytes = proc.stderr.read()
        proc.stderr.close()
        proc.stderr = io.BytesIO(out)
    else:
        out = b""
    return out


class ExpectProc:
    def __init__(self, proc: "subprocess.Popen[bytes]"):
        self.proc = proc
        self.source = assert_io(proc.stderr, name="stderr")
        self.dest = assert_io(proc.stdin, name="stdin")
        self.stdout = proc.stdout

    def expect(self, expected: bytes, prefix: bytes = b"[GNUPG:] GET_") -> None:
        actual = b""
        while not actual.startswith(prefix):
            actual = self.source.read(len(prefix))
            c = b"1"
            while c and c != b"\n":
                c = self.source.read(1)
                if c != b"\r":  # Skip '\r' (windows)
                    actual += c
            if not actual:
                raise ValueError("Unexpected end of source")
        if actual != expected:
            raise ValueError(
                f"Unexpected prompt from gpg:\n{expected.decode()}\n{actual.decode()}"
            )

    def put(self, msg: bytes) -> None:
        self.dest.write(msg)
        self.dest.flush()


@contextmanager
def expect(command: Sequence[str], **kwargs: Any) -> Generator[ExpectProc, None, None]:
    # For the bandit vulnerability suppressed here (nosec):
    # see info close the subprocess import
    with subprocess.Popen(  # nosec
        command,
        stdin=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
        **kwargs,
    ) as proc:
        yield ExpectProc(proc)


def feed_from_any(src: ISource) -> Callable[[IO[bytes]], None]:
    if isinstance(src, bytes):
        return feed_from_str(src)
    if callable(src):
        return feed_from_callable(src)
    return feed_from_stream(src)


def feed_from_str(src: bytes) -> Callable[[IO[bytes]], None]:
    def feed(stdin: IO[bytes]) -> None:
        stdin.write(src)

    return feed


def feed_from_stream(src: IO[bytes]) -> Callable[[IO[bytes]], None]:
    def feed(stdin: IO[bytes]) -> None:
        copyfileobj(src, stdin)

    return feed


def feed_from_callable(src: Callable[[IO[bytes]], None]) -> Callable[[IO[bytes]], None]:
    def feed(stdin: IO[bytes]) -> None:
        src(stdin)

    return feed


def has_fileno(s: Any) -> bool:
    if hasattr(s, "fileno"):
        try:
            s.fileno()
        except Exception:  # pylint: disable=broad-except
            return False
        return True
    return False


def assert_io(stream: Optional[IO[bytes]], name: str = "stdout") -> IO[bytes]:
    if stream is None:
        raise ValueError(f"Unexpected {name} of type NoneType.")
    return stream



import re
from typing import Optional, Union, Tuple, Iterator, IO, Callable
import urllib.request
from urllib.parse import unquote, quote
from urllib.error import HTTPError, URLError
from http.client import HTTPResponse

from .model import KeyInfo, Uid


class KeyserverError(Exception):
    """Error class that displays an error message when the keyserver cannot
    be reached.

    :param action: either "download", "upload", "search".
    :param keyserver: URL of keyserver.
    """

    def __init__(self, action: str, keyserver: str):
        super().__init__(
            f"Key {action} failed because the specified keyserver "
            f"[{keyserver}] could not be reached. The keyserver might be "
            "temporarily unavailable or a wrong URL was provided. "
            "Try connecting to the keyserver using your web browser to "
            "confirm whether the keyserver is available or not."
        )


class KeyserverKeyNotFoundError(Exception):
    """Error class that displays an error message when a given key cannot be
    found on a keyserver.

    :param keyserver: URL of keyserver.
    :param fingerprint: fingerprint of PGP key that could not be found.
    """

    def __init__(self, keyserver: str, fingerprint: str):
        super().__init__(
            f"Key [{fingerprint}] was not found on the keyserver [{keyserver}]."
        )


def parse_search_keyserver(source: IO[bytes]) -> Iterator[KeyInfo]:
    """Extract key information (fingerprint, user ID) from the string returned
    by a search request made via the OpenPGP HTTP Keyserver Protocol. See
    https://tools.ietf.org/html/draft-shaw-openpgp-hkp-00 for details.

    Example of keyserver response to key index search:
    info:1:1
    pub:AA1020B5F0F873830B9AF245AA2BB155CF133521:1:4096:1593675716::
    uid:Alice Smith <alice.smith@example.com>:1593675716::
    """
    try:
        b_line = next(source).rstrip(b"\r\n")
    except StopIteration:
        return
    if not b_line.startswith(b"info:"):
        raise ValueError(
            "Unknown response format from gpg, starting with\n"
            + b_line.decode("utf-8", "replace")
        )
    for b_line in source:
        line = b_line.decode("utf-8", "replace").rstrip("\r\n")
        if not line:
            continue
        try:
            (
                hdr_key,
                fingerprint,
                algorithm,
                key_length,
                creation_date,
                expiration_date,
                _,
            ) = line.split(":")
        except ValueError as e:
            raise ValueError(f"Wrong format for search_key: {line}") from e
        try:
            hdr_uid, uid_str, _, _, _ = (
                next(source).decode("utf-8", "replace").rstrip("\n").split(":")
            )
        except StopIteration as e:
            raise ValueError("Unexpected end of source. Expected 'uid:' line") from e
        if hdr_key != "pub":
            raise ValueError(
                f"Unknown response format from gpg: "
                f"expected 'pub:', got '{hdr_key}'"
            )
        if hdr_uid != "uid":
            raise ValueError(
                f"Unknown response format from gpg: "
                f"expected 'uid:', got '{hdr_uid}'"
            )
        uid = Uid.from_str(unquote(uid_str))
        if uid is None:
            raise ValueError("Could not determine uid from gpg response.")
        yield KeyInfo(
            uid=uid,
            fingerprint=fingerprint,
            key_algorithm=int(algorithm),
            creation_date=creation_date,
            expiration_date=expiration_date or None,
            key_length=int(key_length),
        )


def normalize_fingerprint(fp: str) -> str:
    valid_len = (8, 16, 32, 40)
    s = re.sub(r"^0[xX]", "", re.sub(r"\s", "", fp))
    if len(s) not in valid_len:
        raise ValueError(f"Invalid fingerprint length. Allowed lengths {valid_len}")
    try:
        int(s, 16)
    except ValueError as e:
        raise ValueError(
            "Fingerprint contains invalid characters. Allowed characters: 0-9A-Fa-f"
        ) from e
    return f"0x{s}"


UrlOpener = Callable[[Union[str, urllib.request.Request]], HTTPResponse]


def search_keyserver(
    search_term: str, keyserver: str, url_opener: UrlOpener = urllib.request.urlopen
) -> Iterator[KeyInfo]:
    """Search for a key on a keyserver supporting the OpenPGP HTTP Keyserver
    Protocol, e.g. an SKS keyserver.

    :param search_term: search term. This can be either a key fingerprint,
        key ID or part of the key user ID (e.g. email address).
    :param keyserver: URL of keyserver where to search for keys.
    :param url_opener: function/wrapper through which to make the http request.
    :return: key information summary.
    :raises KeyserverError: error raised if the keyserver does not respond
        (e.g. wrong URL) or does not accept the request for any other reason.
    """
    try:
        queries = {normalize_fingerprint(search_term), search_term}
    except ValueError:
        queries = {search_term}

    for query in queries:
        try:
            with query_keyserver(
                keyserver=keyserver, query=query, op="index", url_opener=url_opener
            ) as response:
                yield from parse_search_keyserver(response)

        except HTTPError as e:
            if e.code != 404:
                raise KeyserverError(action="search", keyserver=keyserver) from e

        except URLError as e:
            raise KeyserverError(action="search", keyserver=keyserver) from e


def download_key(
    fingerprint: str,
    keyserver: str,
    url_opener: UrlOpener = urllib.request.urlopen,
    **kwargs: str,
) -> HTTPResponse:
    """Download key from keyserver to a user's local keyring. Keys can only
    be retrieved via their fingerprint, or key ID. It is recommended to use
    full fingerprints (40 chars hexadecimal string) to avoid key collisions.

    :param fingerprint: fingerprint (40 chars) or key ID (8, 16 or 32 chars) of
        the key to download.
    :param keyserver: URL of keyserver from where to download keys.
    :param kwargs:
    :return: urlopen context manager containing the downloaded PGP key.
    :raises KeyserverError: error raised if the keyserver does not respond
        (e.g. wrong URL) or does not accept the request for any other reason.
    :raises KeyserverKeyNotFoundError: error raised if no key matching the
        specified fingerprint is present on the keyserver.
    """
    try:
        return query_keyserver(
            keyserver=keyserver,
            query=normalize_fingerprint(fingerprint),
            op="get",
            exact="on",
            url_opener=url_opener,
            **kwargs,
        )

    except HTTPError as e:
        if e.code == 404:
            raise KeyserverKeyNotFoundError(keyserver, fingerprint) from None
        raise KeyserverError(action="download", keyserver=keyserver) from e

    except URLError as e:
        raise KeyserverError(action="download", keyserver=keyserver) from e


def query_keyserver(
    keyserver: str,
    query: str,
    op: str,
    url_opener: UrlOpener = urllib.request.urlopen,
    **parameters: str,
) -> HTTPResponse:
    """Send a request to a keyserver supporting the OpenPGP HTTP Keyserver
    Protocol, e.g. to download a key or retrieve information about a key.

    :param keyserver: URL of keyserver to query.
    :param query: search/query term, e.g. a normalized fingerprint or email
        address.
    :param op: operation type to perfrom, e.g. "index" to list keys or "get"
        to download keys.
    :param url_opener: function/wrapper through which to make the http request.
    :param parameters: optional arguments to pass to
    :return: urlopen context manager containing the data returned by the query
        made to the keyserver.
    :raises ValueError: error is raised if keyserver URL is incorrect.
    """
    url = build_host(keyserver)
    query_url = f"{url}/pks/lookup?search={quote(query)}&op={op}&options=mr" + "".join(
        f"?{key}={val}" for key, val in parameters.items()
    )
    return url_opener(query_url)


def split_host(url: str) -> Tuple[Optional[str], str, Optional[str]]:
    match = re.fullmatch(r"(https?://|hkp://)?([^:]+)(:[0-9]+)?", url)
    if not match:
        raise ValueError(f"Invalid URL: {url}")
    scheme, host, port = match.groups()
    return scheme, host, port


def build_host(url: str) -> str:
    scheme, host, port = split_host(url)
    if scheme not in {"http://", "https://"}:
        if port in {":80", ":8080", ":11371"}:
            scheme = "http://"
        else:
            scheme = "https://"
    if port is None:
        if scheme == "https://":
            port = ":443"
        else:
            port = ":80"
    return scheme + host + port


default_opener = urllib.request.build_opener(
    urllib.request.HTTPSHandler(debuglevel=0)
).open


def upload_keys(
    keys_as_ascii_armor: bytes, keyserver: str, url_opener: UrlOpener = default_opener
) -> None:
    """Upload one or more keys specified by their fingerprints to a SKS
    keyserver.

    Following https://tools.ietf.org/html/draft-shaw-openpgp-hkp-00#page-6,
    Section 4

    :param keys_as_ascii_armor: ascii representation of a PGP key.
    :param keyserver: URL of keyserver to upload keys to.
    :param url_opener: function/wrapper through which to make the http request.
    :raises KeyserverError: error raised if the keyserver does not respond or
        is otherwise unreachable (e.g. wrong URL).
    """
    try:
        post(
            url=build_host(keyserver) + "/pks/add",
            data=b"keytext=" + quote(keys_as_ascii_armor).encode(),
            url_opener=url_opener,
        )
    except URLError as e:
        raise KeyserverError(action="upload", keyserver=keyserver) from e


def post(url: str, data: bytes, url_opener: UrlOpener = default_opener) -> bytes:
    """Send a POST request containing "data" to the specified "url".

    :param url: URL where to send the POST request.
    :param data: data (bytes) to send in the POST request.
    :param url_opener: function/wrapper through which to make the http request.
    """
    request = urllib.request.Request(
        url,
        data,
        method="POST",
        headers={"Content-Type": "application/x-www-form-urlencoded"},
    )
    with url_opener(request) as response:
        response_text = response.read()
        return response_text



import io
import os
import unittest
import tempfile
import functools
from unittest import mock
from urllib.error import HTTPError, URLError

from test.test_gpg import (
    CN_UID,
    CN_FINGERPRINT,
    UMLAUTS_UID,
    UMLAUTS_FINGERPRINT,
    CN_PUB_KEY,
    UMLAUTS_PUB_KEY,
    SGT_HARTMANN_PUB_KEY,
    IntegrationTest,
    ENV_WITH_GPG,
    KEY_SERVER,
)

import gpg_lite as gpg
from gpg_lite import keyserver


ENV_KEY_SERVER = os.environ.get("KEY_SERVER", None)
KEY_SERVER = "http://hagrid.hogwarts.org:11371"

# Variables holding different types of errors to simulate non-responding
# keyserver or bad response from keyserver.
_URLError = URLError("timeout")
HTTPError_404 = HTTPError(url=KEY_SERVER, code=404, msg="not found", hdrs={}, fp=None)
HTTPError_401 = HTTPError(
    url=KEY_SERVER, code=401, msg="unauthorized", hdrs={}, fp=None
)


class TestKeyserver(unittest.TestCase):
    def test_parse_search_keyserver(self) -> None:
        response = io.BytesIO(
            b"""info:1:1
pub:CB5A3A5C2C4419BB09EFF879D6D229C51AB5D107:1:2048:1550241679::
uid:Chuck Norris <chuck.norris@roundhouse.gov>:1550241679::

pub:F119E26211119BB09EFF879D6D229C51AB5D107C:1:2048:1550241679::
uid:With Uml%C3%A4uts <with.umlauts@umlauts.info>:1550241679::

pub:DB5A3A5C2C4419BB09EFF879D6D229C51AB5D107:1:4096:1550241679:1850241679:
uid:Chuck Norris <chuck.norris@roundhouse.gov>:1650241679::
"""
        )
        keys = list(keyserver.parse_search_keyserver(response))
        self.assertEqual(
            keys,
            [
                gpg.KeyInfo(
                    uid=CN_UID,
                    fingerprint=CN_FINGERPRINT,
                    key_algorithm=1,
                    key_length=2048,
                    creation_date="1550241679",
                    expiration_date=None,
                ),
                gpg.KeyInfo(
                    uid=UMLAUTS_UID,
                    fingerprint=UMLAUTS_FINGERPRINT,
                    key_algorithm=1,
                    key_length=2048,
                    creation_date="1550241679",
                    expiration_date=None,
                ),
                gpg.KeyInfo(
                    uid=CN_UID,
                    fingerprint="DB5A3A5C2C4419BB09EFF879D6D229C51AB5D107",
                    key_algorithm=1,
                    key_length=4096,
                    creation_date="1550241679",
                    expiration_date="1850241679",
                ),
            ],
        )

    def test_normalize_fingerprint(self) -> None:
        for length in (8, 16, 32, 40):
            fp = "A" * length
            self.assertEqual(keyserver.normalize_fingerprint(fp), f"0x{fp}")
            self.assertEqual(keyserver.normalize_fingerprint(f"0x{fp}"), f"0x{fp}")
        self.assertEqual(keyserver.normalize_fingerprint("123 af  47a"), "0x123af47a")
        with self.assertRaises(ValueError):
            keyserver.normalize_fingerprint("A" * 7)
        with self.assertRaises(ValueError):
            keyserver.normalize_fingerprint("Z" * 8)

    def test_search_keyserver(self) -> None:
        # Check that passing a non-existing keyserver value raises the correct
        # type of error.
        search_keyserver = functools.partial(
            keyserver.search_keyserver,
            search_term=CN_FINGERPRINT,
            keyserver=KEY_SERVER,
        )
        with self.assertRaises(keyserver.KeyserverError):
            list(search_keyserver(url_opener=mock.Mock(side_effect=_URLError)))

        # Check that a missing key does not raises an error, and that any other
        # type of HTTPError does raise an error.
        list(search_keyserver(url_opener=mock.Mock(side_effect=HTTPError_404)))
        with self.assertRaises(keyserver.KeyserverError):
            list(search_keyserver(url_opener=mock.Mock(side_effect=HTTPError_401)))

    def test_download_key(self) -> None:
        # Check that an urllib error of type URLError or HTTPError raises the
        # correct type of error.
        download_key = functools.partial(
            keyserver.download_key,
            fingerprint=CN_FINGERPRINT,
            keyserver=KEY_SERVER,
        )
        with self.assertRaises(keyserver.KeyserverError):
            download_key(url_opener=mock.Mock(side_effect=_URLError))
        with self.assertRaises(keyserver.KeyserverKeyNotFoundError):
            download_key(url_opener=mock.Mock(side_effect=HTTPError_404))
        with self.assertRaises(keyserver.KeyserverError):
            download_key(url_opener=mock.Mock(side_effect=HTTPError_401))

    def test_upload_keys(self) -> None:
        # Check that an urllib error of type URLError raises the correct type
        # of error.
        upload_keys = functools.partial(
            keyserver.upload_keys,
            keys_as_ascii_armor=CN_PUB_KEY,
            keyserver=KEY_SERVER,
        )
        with self.assertRaises(keyserver.KeyserverError):
            upload_keys(url_opener=mock.Mock(side_effect=_URLError))


def to_info(key: gpg.Key) -> gpg.KeyInfo:
    return gpg.KeyInfo(
        uid=key.uids[0],
        fingerprint=key.fingerprint,
        key_algorithm=key.pub_key_algorithm,
        creation_date=key.creation_date,
        expiration_date=key.expiration_date,
        key_length=key.key_length,
    )


@unittest.skipUnless(ENV_KEY_SERVER is not None and ENV_WITH_GPG, "Key Server Tests")
class TestKeyserverIntegration(IntegrationTest):
    def test_send_search(self) -> None:
        if ENV_KEY_SERVER is None:  # To make mypy happy
            return
        self.s.import_file(CN_PUB_KEY)
        self.s.import_file(UMLAUTS_PUB_KEY)
        key_a, key_b = self.s.list_pub_keys()

        self.s.send_keys(key_a.key_id, key_b.key_id, keyserver=ENV_KEY_SERVER)

        keys = keyserver.search_keyserver(
            key_a.uids[0].full_name or "", keyserver=ENV_KEY_SERVER
        )
        self.assertEqual(list(keys), [to_info(key_a)])

        keys = keyserver.search_keyserver(
            key_b.uids[0].full_name or "", keyserver=ENV_KEY_SERVER
        )
        self.assertEqual(list(keys), [to_info(key_b)])

        # By fingerprint
        keys = keyserver.search_keyserver(key_b.fingerprint, keyserver=ENV_KEY_SERVER)
        self.assertEqual(list(keys), [to_info(key_b)])

        # By ID
        keys = keyserver.search_keyserver(key_b.key_id, keyserver=ENV_KEY_SERVER)
        self.assertEqual(list(keys), [to_info(key_b)])

    def test_recv_keys(self) -> None:
        if ENV_KEY_SERVER is None:  # To make mypy happy
            return
        try:
            with tempfile.TemporaryDirectory() as gpg_temp_dir:
                temp_store = gpg.GPGStore(gpg_temp_dir)
                temp_store.import_file(SGT_HARTMANN_PUB_KEY)
                temp_store.import_file(UMLAUTS_PUB_KEY)
                key_a, key_b = temp_store.list_pub_keys(sigs=True)
                self.assertTrue(key_a.signatures or key_b.signatures)
                temp_store.send_keys(
                    key_a.key_id, key_b.key_id, keyserver=ENV_KEY_SERVER
                )
        except FileNotFoundError:
            pass
        self.s.recv_keys(key_a.key_id, key_b.fingerprint, keyserver=ENV_KEY_SERVER)
        keys = set(self.s.list_pub_keys(sigs=True))
        self.assertEqual({key_a, key_b}, keys)



import unittest
import os
import tempfile
import io
from functools import partial
from typing import Tuple, Any, cast, IO, Callable

import gpg_lite as gpg
from gpg_lite import keyserver, store
from gpg_lite.cmd import GPGError
from gpg_lite.model import SignatureValidity

ENV_WITH_GPG = os.environ.get("WITH_GPG", None) == "true"


CN_UID = gpg.Uid(full_name="Chuck Norris", email="chuck.norris@roundhouse.gov")
CN_FINGERPRINT = "CB5A3A5C2C4419BB09EFF879D6D229C51AB5D107"
CN_SIGNATURE = gpg.Signature(
    issuer_uid=CN_UID,
    issuer_key_id=CN_FINGERPRINT[-16:],
    issuer_fingerprint=CN_FINGERPRINT,
    creation_date="1550241679",
    signature_class="13x",
    validity=SignatureValidity.good,
)

AUTHORITY_UID = gpg.Uid(
    full_name="Unit Test Authority (validation key)", email="authority@roundhouse.gov"
)
AUTHORITY_FINGERPRINT = "848FA2707A1AC6060FFE5D253A6500D5C1DE39AC"
AUTHORITY_SIGNATURE = gpg.Signature(
    issuer_uid=AUTHORITY_UID,
    issuer_key_id=AUTHORITY_FINGERPRINT[-16:],
    issuer_fingerprint=None,
    creation_date="1571225229",
    signature_class="13x",
    validity=SignatureValidity.good,
)
AUTHORITY_SIGNATURE_REVOKED_FAKE = gpg.RevocationSignature(
    issuer_uid=AUTHORITY_UID,
    issuer_key_id=AUTHORITY_FINGERPRINT[-16:],
    issuer_fingerprint=None,
    creation_date="1571225230",
    signature_class="30x",
    validity=SignatureValidity.bad,
    reason="00",
    comment="revsig\\nFake revocation",
)

SGT_HARTMANN_UID = gpg.Uid(
    full_name="Gunnery Sgt. Hartmann", email="hartmann@bullshit.gov"
)
SGT_HARTMANN_FINGERPRINT = "DA1A363BC5E11DA73806A8A4E3E4024A0E56221E"
SGT_HARTMANN_SIGNATURE = gpg.Signature(
    issuer_uid=SGT_HARTMANN_UID,
    issuer_key_id=SGT_HARTMANN_FINGERPRINT[-16:],
    issuer_fingerprint=SGT_HARTMANN_FINGERPRINT,
    creation_date="1571232229",
    signature_class="12x",
    validity=SignatureValidity.good,
)
SGT_HARTMANN_SIGNATURE_REVOKED = gpg.RevocationSignature(
    issuer_uid=SGT_HARTMANN_UID,
    issuer_key_id=SGT_HARTMANN_FINGERPRINT[-16:],
    issuer_fingerprint=SGT_HARTMANN_FINGERPRINT,
    creation_date="1571252229",
    signature_class="30x",
    validity=SignatureValidity.good,
    reason="00",
    comment="revsig\\nInvalid email",
)


UMLAUTS_UID = gpg.Uid(full_name="With Umläuts", email="with.umlauts@umlauts.info")
UMLAUTS_FINGERPRINT = "F119E26211119BB09EFF879D6D229C51AB5D107C"
MISSING_PUB_KEY_SIGNATURE = gpg.Signature(
    issuer_uid=None,
    issuer_key_id="2E3E7FE28FE412AE",
    issuer_fingerprint="2866A142A16F54E06EB73CF92E3E7FE28FE412AE",
    creation_date="1573661452",
    signature_class="13x",
    validity=SignatureValidity.no_public_key,
)

KEY_SERVER = "http://hagrid.hogwarts.org:11371"

CN_KEY_PAYLOAD = b"""pub:u:2048:1:D6D229C51AB5D107:1550241679:::u:::scESC::::::23:1571230031:1 http\\x3a//hagrid.hogwarts.org\\x3a11371:
fpr:::::::::CB5A3A5C2C4419BB09EFF879D6D229C51AB5D107:
uid:u::::1550241679::B1E8D4565D26BED47D1EB8BB2B5733AEC1F154B5::Chuck Norris <chuck.norris@roundhouse.gov>::::::::::0:
sig:!::1:D6D229C51AB5D107:1550241679::::Chuck Norris <chuck.norris@roundhouse.gov>:13x::CB5A3A5C2C4419BB09EFF879D6D229C51AB5D107:::8:
sig:!::1:3A6500D5C1DE39AC:1571225229::::Unit Test Authority (validation key) <authority@roundhouse.gov>:13x:::::2:
sig:?::1:2E3E7FE28FE412AE:1573661452::::[User ID not found]:13x::2866A142A16F54E06EB73CF92E3E7FE28FE412AE:::8:
sig:!::1:E3E4024A0E56221E:1571232229::::Gunnery Sgt. Hartmann <hartmann@bullshit.gov>:12x::DA1A363BC5E11DA73806A8A4E3E4024A0E56221E:::10:
rev:!::1:E3E4024A0E56221E:1571252229::::Gunnery Sgt. Hartmann <hartmann@bullshit.gov>:30x,00::DA1A363BC5E11DA73806A8A4E3E4024A0E56221E:::8:::::revsig\\nInvalid email:
rev:-::1:3A6500D5C1DE39AC:1571225230::::Unit Test Authority (validation key) <authority@roundhouse.gov>:30x,00:::::2:::::revsig\\nFake revocation
sub:u:2048:1:D892C41917B20115:1550241679::::::e::::::23:
fpr:::::::::55C5314BB9EFD19AE7CC4774D892C41917B20115:
sig:!::1:D6D229C51AB5D107:1550241679::::Chuck Norris <chuck.norris@roundhouse.gov>:13x::CB5A3A5C2C4419BB09EFF879D6D229C51AB5D107:::8:
"""
CN_KEY = gpg.Key(
    key_id=CN_FINGERPRINT[-16:],
    fingerprint=CN_FINGERPRINT,
    validity=gpg.Validity.ultimately_valid,
    key_length=2048,
    pub_key_algorithm=1,
    creation_date="1550241679",
    signatures=(
        CN_SIGNATURE,
        AUTHORITY_SIGNATURE,
        MISSING_PUB_KEY_SIGNATURE,
        SGT_HARTMANN_SIGNATURE,
        SGT_HARTMANN_SIGNATURE_REVOKED,
        AUTHORITY_SIGNATURE_REVOKED_FAKE,
    ),
    key_capabilities={
        gpg.KeyCapability.sign,
        gpg.KeyCapability.certify,
    },
    combined_key_capabilities={
        gpg.KeyCapability.sign,
        gpg.KeyCapability.certify,
        gpg.KeyCapability.encrypt,
    },
    disabled=False,
    key_type=gpg.KeyType.public,
    uids=(CN_UID,),
    owner_trust="u",
    sub_keys=(
        gpg.SubKey(
            key_type=gpg.KeyType.public,
            key_id="D892C41917B20115",
            fingerprint="55C5314BB9EFD19AE7CC4774D892C41917B20115",
            validity=gpg.Validity.ultimately_valid,
            key_length=2048,
            pub_key_algorithm=1,
            creation_date="1550241679",
            signatures=(CN_SIGNATURE,),
            key_capabilities={gpg.KeyCapability.encrypt},
        ),
    ),
    origin=KEY_SERVER,
)

AUTHORITY_KEY_PAYLOAD = b"""pub:-:4096:1:3A6500D5C1DE39AC:1565257866:::-:::scESC::::::23::0:
fpr:::::::::848FA2707A1AC6060FFE5D253A6500D5C1DE39AC:
uid:-::::1565257866::4D51176EBCDF1F58C3D7E0F76625B168A184BB8E::Unit Test Authority (validation key) <authority@roundhouse.gov>::::::::::0:
sig:!::1:3A6500D5C1DE39AC:1571225229::::Unit Test Authority (validation key) <authority@roundhouse.gov>:13x:::::2:
sub:-:4096:1:A4C037FA518B5C4F:1565257866::::::e::::::23:
fpr:::::::::3AC9B4462D255449989A3D6EA4C037FA518B5C4F:
sig:!::1:3A6500D5C1DE39AC:1571225229::::Unit Test Authority (validation key) <authority@roundhouse.gov>:13x:::::2:
"""
AUTHORITY_KEY = gpg.Key(
    key_id=AUTHORITY_FINGERPRINT[-16:],
    fingerprint=AUTHORITY_FINGERPRINT,
    validity=gpg.Validity.unknown,
    key_length=4096,
    pub_key_algorithm=1,
    creation_date="1565257866",
    signatures=(AUTHORITY_SIGNATURE,),
    key_capabilities={
        gpg.KeyCapability.sign,
        gpg.KeyCapability.certify,
    },
    combined_key_capabilities={
        gpg.KeyCapability.sign,
        gpg.KeyCapability.certify,
        gpg.KeyCapability.encrypt,
    },
    disabled=False,
    key_type=gpg.KeyType.public,
    uids=(AUTHORITY_UID,),
    owner_trust="-",
    sub_keys=(
        gpg.SubKey(
            key_type=gpg.KeyType.public,
            key_id="A4C037FA518B5C4F",
            fingerprint="3AC9B4462D255449989A3D6EA4C037FA518B5C4F",
            validity=gpg.Validity.unknown,
            key_length=4096,
            pub_key_algorithm=1,
            creation_date="1565257866",
            signatures=(AUTHORITY_SIGNATURE,),
            key_capabilities={gpg.KeyCapability.encrypt},
        ),
    ),
    origin=None,
)


class TestUtils(unittest.TestCase):
    def test_split_url(self) -> None:
        self.assertEqual(
            keyserver.split_host("http://hagrid.hogwarts.org:11371"),
            ("http://", "hagrid.hogwarts.org", ":11371"),
        )
        self.assertEqual(
            keyserver.split_host("hkp://hagrid.hogwarts.org:11371"),
            ("hkp://", "hagrid.hogwarts.org", ":11371"),
        )
        self.assertEqual(
            keyserver.split_host("hagrid.hogwarts.org:11371"),
            (None, "hagrid.hogwarts.org", ":11371"),
        )
        self.assertEqual(
            keyserver.split_host("http://hagrid.hogwarts.org"),
            ("http://", "hagrid.hogwarts.org", None),
        )
        self.assertEqual(
            keyserver.split_host("hagrid.hogwarts.org"),
            (None, "hagrid.hogwarts.org", None),
        )
        with self.assertRaises(ValueError):
            keyserver.split_host("hagrid.hogwar:ts.org")

    def test_uid_missging_both_fields(self) -> None:
        with self.assertRaises(ValueError):
            gpg.Uid(full_name=None, email=None)

    def test_uid_from_str(self) -> None:
        self.assertEqual(gpg.Uid.from_str("[User ID not found]"), None)
        self.assertEqual(gpg.Uid.from_str("[User-ID nicht gefunden]"), None)
        self.assertEqual(gpg.Uid.from_str("[Identité introuvable]"), None)
        self.assertEqual(
            gpg.Uid.from_str("Chuck Norris <chucknorris@roundhouse.gov>"),
            gpg.Uid(full_name="Chuck Norris", email="chucknorris@roundhouse.gov"),
        )
        self.assertEqual(
            gpg.Uid.from_str(
                "Chuck Norris (Chucks's test PGP key) " "<chucknorris@roundhouse.gov>"
            ),
            gpg.Uid(
                full_name="Chuck Norris (Chucks's test PGP key)",
                email="chucknorris@roundhouse.gov",
            ),
        )
        self.assertEqual(
            gpg.Uid.from_str("Chuck Norris"),
            gpg.Uid(full_name="Chuck Norris", email=None),
        )
        self.assertEqual(
            gpg.Uid.from_str("chucknorris@roundhouse.gov"),
            gpg.Uid(full_name=None, email="chucknorris@roundhouse.gov"),
        )

    def test_uid_repr(self) -> None:
        cases = (
            (
                "Chuck Norris (Chucks's test PGP key)",
                "chucknorris@roundhouse.gov",
                "Chuck Norris (Chucks's test PGP key) <chucknorris@roundhouse.gov>",
            ),
            (
                "Chuck Norris",
                "chucknorris@roundhouse.gov",
                "Chuck Norris <chucknorris@roundhouse.gov>",
            ),
            ("Chuck Norris", None, "Chuck Norris"),
            (None, "chucknorris@roundhouse.gov", "chucknorris@roundhouse.gov"),
        )
        for name, email, expected in cases:
            self.assertEqual(repr(gpg.Uid(full_name=name, email=email)), expected)


class TestGPG(unittest.TestCase):
    def test_parse_keys(self) -> None:
        # Verify parsing of GnuPG output for single and multiple key(s).
        for keys, payload in (
            ([CN_KEY], CN_KEY_PAYLOAD),
            ([AUTHORITY_KEY], AUTHORITY_KEY_PAYLOAD),
            ([CN_KEY, AUTHORITY_KEY], CN_KEY_PAYLOAD + AUTHORITY_KEY_PAYLOAD),
        ):
            parsed_keys = store.parse_keys(
                io.BytesIO(b"tru::1:1571400791:0:3:1:5\n" + payload)
            )
            self.assertEqual(parsed_keys, keys)

    def test_parse_keys_revocation_sig(self) -> None:
        (parsed_key,) = store.parse_keys(
            io.BytesIO(b"tru::1:1571400791:0:3:1:5\n" + CN_KEY_PAYLOAD)
        )
        self.assertEqual(parsed_key, CN_KEY)

        # Verify that valid signatures are detected correctly. The following
        # signatures are invalid and should not be listed:
        #  * MISSING_PUB_KEY_SIGNATURE: signature that could not be verified
        #       because of missing public key in the user's keyring.
        #  * SGT_HARTMANN_SIGNATURE: revoked by SGT_HARTMANN_SIGNATURE_REVOKED
        #  * AUTHORITY_SIGNATURE_REVOKED_FAKE: this does actually not
        #       revoke AUTHORITY_SIGNATURE because the revocation signature
        #       is invalid. Therefore AUTHORITY_SIGNATURE must still be listed
        #       as a valid key.
        self.assertEqual(
            parsed_key.valid_signatures, [CN_SIGNATURE, AUTHORITY_SIGNATURE]
        )

    def test_extract_key_id(self) -> None:
        partial_msg = io.BytesIO(b"\x84\x8c\x03\xf3\rL3@z?c\x01\x03\xff]\xdf")
        pos = partial_msg.tell()
        (key_id,) = gpg.extract_key_id(partial_msg)
        self.assertEqual(key_id, "F30D4C33407A3F63")
        self.assertEqual(partial_msg.tell(), pos)

        wrong_msg = io.BytesIO(b"\x04\x8c\x03\xf3\rL3@z?c\x01\x03\xff]\xdf")
        with self.assertRaises(gpg.GPGError):
            list(gpg.extract_key_id(wrong_msg))

        partial_msg = io.BytesIO(
            b"\x85\x01\x0c\x03\xd8\x92\xc4\x19\x17\xb2\x01\x15\x01\x08"
            b"\x00\xafZ\xd8\x83\xba"
            + 248 * b"\000"
            + b"c\x1d\x16\x85\x02\x0c\x032\xf9\xe2\xf8\rR!^\x01\x10\x00\x97c"
            + 504 * b"\000"
            + b"\x1a\xa2\xf2&\xc0\xe5\x85\x02\x0c\x03\xcc\xfa\xa7\xf4\x12\xb0m"
            b"\x83\x01\x10"
        )

        key_ids = list(gpg.extract_key_id(partial_msg))
        self.assertEqual(
            key_ids, ["D892C41917B20115", "32F9E2F80D52215E", "CCFAA7F412B06D83"]
        )

    def test_extract_key_id_ascii_armored(self) -> None:
        partial_msg = io.BytesIO(b"-----BEGIN PGP MESSAGE-----\r\n\r\n")
        with self.assertRaises(gpg.GPGError) as cm:
            list(gpg.extract_key_id(partial_msg))
        self.assertEqual(
            format(cm.exception), "ASCII armored encrypted data is not supported"
        )

    def test_extract_key_id_from_sig(self) -> None:
        sig = (
            b"\x89\x023\x04\x00\x01\x08\x00\x1d\x16!\x04\xa8\x05h*\x13\xa1\x08\xca"
            b"\xab\x07\xef\xd0k\xe5(\xb1&\xd3\xe13\x05\x02_\x05\xcfP\x00\n\t\x10k"
            b"\xe5(\xb1&\xd3\xe13\xf8\x8a\x10\x00\x97\xb8\x0bWU\x05B\xd5\x17\xba\xd2"
            b"\x9b\xe4o\xd4`\x86Id\x86\xc4m\xa6=\xf2\xf3\xc5\xec\x8b\x16?Q\xfe\x06"
            b"\xf9\xdc`\xfa\xe4\x9d~)\xcaYx\x16\x06t@\xc0\xc0\xe4\xd5\x94\x1a\x19"
            b"\xa5yL.r\xba\x01-\xe8\xc6\xaf \xa8\xc9\xbef\x0b:C\xa4\x83Ahn\x85R\xbe"
            b"\x81\x02\xb14\xce\x89\xef4\xa4\xb5\x96\xcc\x81t/x\x16_\xb1|\xcbb+3"
            b"\xf4`w_\x93\x8cl\xa0\x95$<&\x8fu<\n\xd1-9^\\\x98\x04\xa0F\x9fE\xeb{"
            b"\xc4\xda\x9f\x9f\xf0\xc9#>\xba\xb5\xa8\x99Od\xcdb\x9en\xa0\xb2\x1c\xd2"
            b"\xe8\xa8\x03k\x04o\xd5\x08\xb7W\x19UL\xa4\x14P\xba\x10}XS\x18\x86M"
            b"\xa3k\xdc\x0fO&\xe6J-\xca\x13\xac\xe8\xacb\xc5\x81x\xe3\xf0\x0b\xe1["
            b"\x1f\xc7\xd1b\x89\xb7\xe3E\x06\xa7n\x83\xe8\xc0.\x96\xd8\x19\x9d\xc3J"
            b"\xfcR\xc1%\x960\x95Y\n\x98`\xb1C(KAwO\xf6\xf5\xe1\xdcCk\x96j\xbbC\xd3?"
            b"\x05\x07r\xc1#\x97\x1d\x15~L\xb8HQ\xb4|=\x8dG\x987wN\xc8\xbfm\x889\xd5"
            b"\x9c\xfd\x90@#C\xef\xcf\x12\r\xc1\xdd\x1e\xd2\xc3\x97[\xad\xcc\x86\xbc"
            b"\xa2\xc6R\xce\xf6L\x8f5;\xe0\xc4|\xe2\xeb\xfe\xe5\xb0\xa2F\xc2\x8eX"
            b"\xab \xe7\xe3\xf3\xe9SN\x16K\x99\xb9S}\xa7\x02\xaa\xd6\x9c\xb3\xb8"
            b"\xcelO\xd3!/2r|\x06\x10\x88$L\x99\x15 \xff\xa7\x812\xb6\xf6I\xba\x10"
            b"\xfd\x94\xb6\x05:`*\xa3\xd0\xfb\xf8~\x0f\xedPx\t\x1fq\xe0\x8e\xf0\xcb"
            b"\xf1\xab\xa0yF\xc7iL\xba\xb7e\xfa\xba\xfb\x8e\xe5w\x8bF\xd3\x91\xc7G4"
            b"\x9c]\xcc\xb8\x10\x0c\xdc\x81\xc0_\x1e\x83c=x2\xd6\xa9:s\xc5\xf0\x04#"
            b"\x0c`\xa2\x0c\x95&\x08\xf5:\x12\xe0\x8a\xb6\x86\x9dE\xea\xa9\xd6\xff"
            b"\xe9~\x91_|\x14.\xcb\x14\xdd\xb9\tR\x1a)\x04\xaf\xf6\xb7\xd4(&%\xc4"
            b"\xe9\xea\xac\x00\x17\xe3Nf\xd6\xd1\xa0\xb9\x99\xe3[\xa1O?a\xd7\x06\x9fc"
        )
        key_id = gpg.extract_key_id_from_sig(sig)
        self.assertEqual(key_id, "A805682A13A108CAAB07EFD06BE528B126D3E133")
        sig = b"""-----BEGIN PGP SIGNATURE-----

iQIzBAABCAAdFiEEqAVoKhOhCMqrB+/Qa+UosSbT4TMFAl8YCiwACgkQa+UosSbT
4TMEvQ/5AaSlYmpQK1MWQbNHiL4AMl55GfVlSnQBCpk3lsltOnycEVxP67jawpd7
2DSFBzcF5wRUiMo85R9YkNxBjFwfU2xlT738ZWFQIoOvvwyGJgOlgngZ7+BEGEqP
130pEIGeDVt6kzVuwqB09SZCATGR0qnkB/84+pFimN0tiFkFWHXLHgJ++uMO2Pg+
j45HbZzYh0qqKfY+X9lP5JIuCsnFQv5AOgnUFfy1vKNZd6OGCZuxCla5M+RpOcFD
F3w004RK2p3SNS7ikoATRSDnSHwiWDay8WesmVkOXQurevEONGdvUY91ZW41wA5T
HR3qHWZyg4fWy4G2azCbGA3AEnd2m+sWdvptpbL7InxutLET5hLEXNsnL1Ixiqr/
DViv+hBaixAT0LRCdci3G/FfWAC38zZjo/A6JLpCGaqTuWl1OWkDPxQXPNYT4Wgk
Y9YJFrgF1MQ2BrUuYzrC+ClEvO/TBsVe2r9l4ElZs3rZBH1vOxyCyuAHreNw46R6
MfwW9lYob56DUTRBVrhZI46/LQD/dlwtYedN5p/CNpzavuzQHAmJD1G22WZHwTqV
SL/bsOqwOqJP1BpQloAR+jEnP6jPbQAmPEn3xU+3IPWjKFe95QUj3PIeeSqst2Mt
xERXJ/PqFoAktjU/mhNeDHovXdB+7378YkY3NHNSlFytYG0D5jo=
=qg97
-----END PGP SIGNATURE-----
"""
        key_id = gpg.extract_key_id_from_sig(sig)
        self.assertEqual(key_id, "A805682A13A108CAAB07EFD06BE528B126D3E133")
        key_id = gpg.extract_key_id_from_sig(sig.replace(b"\n", b"\r\n"))
        self.assertEqual(key_id, "A805682A13A108CAAB07EFD06BE528B126D3E133")

        # Signature generated with gpg 2.0.22
        sig = (
            b"\x89\x02\x1c\x04\x00\x01\x02\x00\x06\x05\x02a`,\xa6\x00\n\t\x10\xab"
            b"\xd0chZI\x86\x83\xa2\xeb\x10\x00\x8e\xe6\x15\x0e\x9cf\xd6t\x07"
            b"?f\x81\x100\x88\xa7\x94\x18\xf1j\x9c\xcc]\xb1D\x13\xd3A?"
            b"\xb5(c\x85\xa9\xd86\xdc\x1fm\x99\xbb:D\x8b\x08D/\x85\xea"
            b"\xe5\x17\xc5'\xca_8\x94\x00\x07\xb3\x96\xe6dJ\x0e\xed\xb7F,"
            b"/\xaa\xb5\xaeB\x99\x17\xeb\xc9\xf4\xf1\xa5\xb0\x03\xe2l\x08\xf7\x8d\xfe"
            b" \xee\x04x\xe7h\xbdv<\x98\x8d=\x84\xa6\xc8q\xd0wVJ"
            b'\x1c\x83\x85\xf4o\xd0$ \xa543\xcd\xe5\x05\xc9\x10\xac\x8f"F'
            b"\x1b\xfd\xa6<dKT\xd2\x13\x9e\xe1\x9bd\xc6Q\xb0n\xd2`Z"
            b"\xda\x11\xdd\xe2\x95\xb6uY\xff\x8bA\xb1\xdd.\x0f|\xb1\x9d<c"
            b"\xd0\x1f\xf3\x9f\xcdI\xb4\xb7\x05\xba\x0f\xc5\x1d\x9e\x0b\xeaT\xc76\xaa"
            b"\x05f\xe3\xbe\xf2\xd6U\xc9sE?\xf9\xbd\xb2p0n\x9a@\xec"
            b".\x9c'\xdb\x86{\x96?\x13|W\xbew[G\x94<\xc0\x8d\""
            b"N\xe1C2\xed=\r\x8b\xb9]b\xb7q\xc6\xba\x8ev/8\xa1"
            b"\xf6\xef\xc5J\xdb\xd7\xd2v\xeb\xbf\xd7\x18\x89\x95\xe3=`\xbf\x0c\xa8"
            b"\xb4\xab\x0c\xcb\x07\xe2\xcc\xfa\xed\xe4MI\xec\x08\x94\n\x92\xd7\xfc\xbb"
            b'\xc4\x18Gf\x9c{v\xbc\xb2"\xaaa.\xab\x03\xbb\xf3@\xcbg'
            b"\xb5m\xe1\r`\xa8/\x98\xa6\x843\xd1\x80\xc3\xcdj\x9b6\x8c4"
            b"G\xfc\xfe\x807S*\xf9Ts\xd2^\xb9\xa1\xe5(]V\xbf$"
            b"\xf0\x89\xb3\xbaX\xcdL-\xa0\xca\xdc\tZA\x8e\xad).Q\xf2"
            b"*\xc2t\x93\x8f\xff\xd6\xce\xddq\x14\xe1+gJ(V\xdd7\xd3"
            b"\xed\x01\xcd\x9d\xbdPq\x87\x88\xb3\xc1\x8e\xed\xdc\x9bl\x0e\xb5\xbfI"
            b"\x8c\x0b\xc2\x9cF\xbeib\x9av\xee~\xe59OvJ\xdc+0"
            b"N\xd4\xd6\x0f,V\xda\xb9\x0b\xc0J\xfd\xb7l\xd1<\xd2\x0c\xd5\xba"
            b"|9\xe0\x1fj19\xbc3\x93\x92t%\x10/<\xfc'\xa7\xe7"
            b"\xf0\x8d\xad\x0c\xab\x9f`.\xdf@K\xef\xf2A\xca]+'\x9f\xab"
            b"\xc7\xcd\xd9\x0b\xb4\x87:<\xef\x9a\xa4\x97XEvDi\x1a\x1a!"
            b"\xf1\xa0\xf2"
        )
        key_id = gpg.extract_key_id_from_sig(sig)
        self.assertEqual(key_id, "ABD063685A498683")

    def test_capture_fpr(self) -> None:
        b = b"""
[GNUPG:] NEWSIG
gpg: Signature made Tue 10 Mar 2020 11:29:27 CET
gpg:                using RSA key AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
[GNUPG:] KEY_CONSIDERED AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA 0
[GNUPG:] SIG_ID PaAuiNMUyUDRMsctXpV+7rX+0CI 2020-03-10 1583836167
[GNUPG:] KEY_CONSIDERED AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA 0
[GNUPG:] GOODSIG 066E042415D8FD9C Chuck Norris <chucknorris@roundhouse.gov>
gpg: Good signature from "Chuck Norris <chucknorris@roundhouse.gov>" [undefined]
[GNUPG:] VALIDSIG AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA 2020-03-10 1583836167 0 4 0 1 8 00 AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
[GNUPG:] TRUST_UNDEFINED 0 pgp
gpg: WARNING: This key is not certified with a trusted signature!
gpg:          There is no indication that the signature belongs to the owner.
"""
        self.assertEqual(
            store.capture_fpr(b), ["AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"]
        )
        b = b"""
gpg: Signature made Mi 12 Feb 2020 15:34:46 CET
gpg: WARNING: This key is not certified with a trusted signature!
gpg:          There is no indication that the signature belongs to the owner.
"""
        self.assertEqual(store.capture_fpr(b), [])
        b = b"""
[GNUPG:] NEWSIG
gpg: Signature made Tue 10 Mar 2020 11:29:27 CET
gpg:                using RSA key AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
[GNUPG:] KEY_CONSIDERED AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA 0
[GNUPG:] SIG_ID PaAuiNMUyUDRMsctXpV+7rX+0CI 2020-03-10 1583836167
[GNUPG:] KEY_CONSIDERED AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA 0
[GNUPG:] GOODSIG 066E042415D8FD9C Chuck Norris <chucknorris@roundhouse.gov>
gpg: Good signature from "Chuck Norris <chucknorris@roundhouse.gov>" [undefined]
[GNUPG:] VALIDSIG AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA 2020-03-10 1583836167 0 4 0 1 8 00 AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
[GNUPG:] TRUST_UNDEFINED 0 pgp
gpg: WARNING: This key is not certified with a trusted signature!
gpg:          There is no indication that the signature belongs to the owner.
[GNUPG:] NEWSIG
gpg: Signature made Tue 10 Mar 2020 11:29:27 CET
gpg:                using RSA key BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB
[GNUPG:] KEY_CONSIDERED BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB 0
[GNUPG:] SIG_ID PaAuiNMUyUDRMsctXpV+7rX+0CI 2020-03-10 1583836167
[GNUPG:] KEY_CONSIDERED BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB 0
[GNUPG:] GOODSIG 066E042415D8FD9C Gunnery Sgt. Hartmann <hartmann@bullshit.gov>
gpg: Good signature from "Gunnery Sgt. Hartmann <hartmann@bullshit.gov" [undefined]
[GNUPG:] VALIDSIG BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB 2020-03-10 1583836167 0 4 0 1 8 00 BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB
[GNUPG:] TRUST_UNDEFINED 0 pgp
gpg: WARNING: This key is not certified with a trusted signature!
gpg:          There is no indication that the signature belongs to the owner.
"""
        self.assertEqual(
            store.capture_fpr(b),
            [
                "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
                "BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB",
            ],
        )

    def test_assert_mdc(self) -> None:
        b = b"""
[GNUPG:] ENC_TO AAAAAAAAAAAAA 1 0
[GNUPG:] KEY_CONSIDERED AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA 0
[GNUPG:] KEY_CONSIDERED AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA 0
[GNUPG:] DECRYPTION_KEY BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA u
[GNUPG:] KEY_CONSIDERED AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA 0
gpg: encrypted with 4096-bit RSA key, ID AAAAAAAAAAAAA, created 2019-02-18
      "Chuck Norris <chucknorris@roundhouse.gov>"
[GNUPG:] BEGIN_DECRYPTION
[GNUPG:] DECRYPTION_COMPLIANCE_MODE 23
[GNUPG:] DECRYPTION_INFO 2 9
[GNUPG:] PLAINTEXT 62 1628663575 test.txt
[GNUPG:] PLAINTEXT_LENGTH 14
[GNUPG:] NEWSIG
gpg: Signature made Wed 11 Aug 2021 08:32:55 CEST
gpg:                using RSA key AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
[GNUPG:] KEY_CONSIDERED AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA 0
[GNUPG:] SIG_ID txTzNz6Yi2We04M0MWGw3pDp73w 2021-08-11 1628663575
[GNUPG:] KEY_CONSIDERED AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA 0
[GNUPG:] GOODSIG 066E042415D8FD9C Chuck Norris <chucknorris@roundhouse.gov>
gpg: Good signature from "Chuck Norris <chucknorris@roundhouse.gov>" [ultimate]
[GNUPG:] VALIDSIG AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA 2021-08-11 1628663575 0 4 0 1 10 00 AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
[GNUPG:] KEY_CONSIDERED AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA 0
[GNUPG:] TRUST_ULTIMATE 0 pgp
[GNUPG:] VERIFICATION_COMPLIANCE_MODE 23
[GNUPG:] DECRYPTION_OKAY
[GNUPG:] GOODMDC
[GNUPG:] END_DECRYPTION
"""
        store.assert_mdc(b)

        with self.assertRaises(GPGError) as cm:
            store.assert_mdc(
                b.replace(
                    b"[GNUPG:] DECRYPTION_OKAY\n[GNUPG:] GOODMDC",
                    b"gpg: WARNING: message was not integrity protected\n"
                    b"gpg: decryption forced to fail!\n[GNUPG:] DECRYPTION_FAILED",
                )
            )
        self.assertEqual(str(cm.exception), "Message integrity could not be verified.")


class IntegrationTest(unittest.TestCase):
    def _setup_key(self) -> gpg.Key:
        self.s.import_file(CN_SEC_KEY, passphrase="test")
        self.s.import_file(CN_PUB_KEY)
        (key,) = self.s.list_sec_keys()
        return key

    def setUp(self) -> None:
        # pylint: disable=consider-using-with
        gpg_homedir = tempfile.TemporaryDirectory()
        self.gpg_homedir = gpg_homedir
        d = gpg_homedir.__enter__()
        self.s = gpg.GPGStore(d)

    def tearDown(self) -> None:
        try:
            self.gpg_homedir.__exit__(None, None, None)
        except FileNotFoundError:
            pass


@unittest.skipUnless(ENV_WITH_GPG, "Integration tests")
class TestGPGIntegration(IntegrationTest):
    def test_revoke_key(self) -> None:
        if self.s._version < (2, 2, 0):  # pylint: disable=protected-access
            self.skipTest("Key revocation not supported for gnupg < 2.2.0")
        key = self._setup_key()
        self.assertTrue(key.validity != gpg.Validity.revoked)
        self.s.revoke_key(key.fingerprint, passphrase="test")
        (key,) = self.s.list_sec_keys()
        self.assertEqual(key.validity, gpg.Validity.revoked)

    def test_list_keys(self) -> None:
        self._setup_key()
        self.s.import_file(UMLAUTS_PUB_KEY)

        def extract_keys(**kwargs: Any) -> Tuple[gpg.Key, gpg.Key]:
            keys = self.s.list_pub_keys(**kwargs)
            if keys[0].uids == (CN_UID,):
                key_1, key_2 = keys
            else:
                key_2, key_1 = keys
            return key_1, key_2

        pub_key, umlauts_key = extract_keys()

        self.assertEqual(pub_key.uids, (CN_UID,))
        self.assertEqual(pub_key.key_type, gpg.KeyType.public)
        self.assertFalse(pub_key.signatures)
        self.assertEqual(umlauts_key.uids, (UMLAUTS_UID,))
        self.assertEqual(umlauts_key.key_type, gpg.KeyType.public)
        self.assertFalse(umlauts_key.signatures)

        pub_key, umlauts_key = extract_keys(sigs=True)
        self.assertEqual(pub_key.uids, (CN_UID,))
        self.assertEqual(pub_key.key_type, gpg.KeyType.public)
        self.assertTrue(pub_key.signatures)
        self.assertEqual(umlauts_key.uids, (UMLAUTS_UID,))
        self.assertEqual(umlauts_key.key_type, gpg.KeyType.public)
        self.assertTrue(umlauts_key.signatures)

        keys = self.s.list_pub_keys(search_terms=(pub_key.fingerprint,), sigs=True)
        self.assertEqual(keys, [pub_key])

        (sec_key,) = self.s.list_sec_keys()
        self.assertEqual(sec_key.uids, (CN_UID,))
        self.assertEqual(sec_key.key_type, gpg.KeyType.secret)

        keys = self.s.list_pub_keys(search_terms=(), sigs=True)
        self.assertEqual(keys, [])

    def test_gen_key(self) -> None:
        self.s.gen_key(
            full_name=cast(str, CN_UID.full_name),
            email=cast(str, CN_UID.email),
            passphrase="test",
            key_length=1024,
        )
        (pub_key,) = self.s.list_pub_keys()
        self.assertEqual(pub_key.uids, (CN_UID,))
        self.assertEqual(pub_key.key_length, 1024)
        self.assertEqual(pub_key.key_type, gpg.KeyType.public)
        (sec_key,) = self.s.list_sec_keys()
        self.assertEqual(sec_key.uids, (CN_UID,))
        self.assertEqual(sec_key.key_length, 1024)
        self.assertEqual(sec_key.key_type, gpg.KeyType.secret)

    def test_gen_key_unprotected_ecc(self) -> None:
        if self.s._version < (2, 2, 0):  # pylint: disable=protected-access
            self.skipTest("ECC keys not supported for gnupg < 2.2.0")
        self.s.gen_key(
            full_name=cast(str, CN_UID.full_name),
            email=cast(str, CN_UID.email),
            passphrase=None,
            key_length=1024,
            key_type="eddsa",
            subkey_type="ecdh",
            key_curve="Ed25519",
            subkey_curve="Curve25519",
        )
        # key length does not seem to be respected by ECC keys
        (pub_key,) = self.s.list_pub_keys()
        self.assertEqual(pub_key.uids, (CN_UID,))
        self.assertEqual(pub_key.key_type, gpg.KeyType.public)
        (sec_key,) = self.s.list_sec_keys()
        self.assertEqual(sec_key.uids, (CN_UID,))
        self.assertEqual(sec_key.key_type, gpg.KeyType.secret)

    def test_decrypt(self) -> None:
        message = b"a"
        self.s.import_file(CN_SEC_KEY, passphrase="test")
        decrypted = io.BytesIO()

        def capture_output(stream: IO[bytes]) -> None:
            decrypted.write(stream.read())
            decrypted.seek(0)

        with self.assertRaises(BaseException):
            self.s.decrypt(
                source=io.BytesIO(ENCRYPTED_MESSAGE),
                output=capture_output,
                passphrase="wrongpass",
            )
        self.s.decrypt(
            source=io.BytesIO(ENCRYPTED_MESSAGE),
            output=capture_output,
            passphrase="test",
        )
        self.assertEqual(decrypted.read(), message)

        with self.subTest("Missing MDC"):
            decrypted.seek(0)
            decrypted.truncate()
            encrypted_msg_without_mdc = (
                b"\x84\x8c\x03\xf3\rL3@z?c\x01\x03\xfe=\x97\x01kJP"
                b"\x90H\x06\x9a\xa2\x17!\xa8\xf9$n\xcf/\xad\xcb\xac\xc3;Y\xe9"
                b"\xebw\x0c7^Z<\xces{\xfe\x9a*`\xcf\x01\x9d\xbcZ\r"
                b"\xf4\x1ae0a3\x97 p\xc8\x9afa\xbd\x93\xea\xa3\xa4\xb9{"
                b'\x18"\x10%\x9cZ\xea\xa8,,\xd7zt\xcf\xe3\xce.4\xe16'
                b"\x01\xd0'\xab\x050<3\xc3xA\x13\xf3\x83\x83\xa2}\xc3X\xc5"
                b"\xb1vE\xdeV\xd5of\x8c\xd1*\x04\xa03\xb6\xa5\xc0\xaf\x14~"
                b"\xa3\x8c\xc9&\x18\xc5\xf4\xb8\x04\x99\x8a\t\x82f\xc2\xe1\xaf\xad\x85D"
                b"\xf0\xfe2\xc5\x865\xb8m\xa3R\xd0\xfa\x13\xba*\x85\x0c\x9f\xbdS"
                b"!\xbf"
            )
            with self.assertRaises(GPGError) as cm:
                self.s.decrypt(
                    source=io.BytesIO(encrypted_msg_without_mdc),
                    output=capture_output,
                    passphrase="test",
                )
            self.assertEqual(
                str(cm.exception), "Message integrity could not be verified."
            )

    def test_encrypt_decrypt(self) -> None:
        key = self._setup_key()
        message = b"a"
        decrypted = io.BytesIO()

        def capture_output(stream: IO[bytes]) -> None:
            decrypted.write(stream.read())
            decrypted.seek(0)

        decrypt = cast(
            Callable[[IO[bytes]], None],
            partial(self.s.decrypt, output=capture_output, passphrase="test"),
        )
        self.s.encrypt(
            source=message,
            output=decrypt,
            recipients=(key.fingerprint,),
            trust_model=gpg.TrustModel.always,
        )
        self.assertEqual(decrypted.read(), message)

    def test_encrypt_decrypt_compression(self) -> None:
        key = self._setup_key()
        message = b"a" * 50
        compressed = io.BytesIO()
        not_compressed = io.BytesIO()

        def capture_output(stream: IO[bytes], buffer: IO[bytes]) -> None:
            buffer.write(stream.read())
            buffer.seek(0)

        self.s.encrypt(
            source=message,
            output=partial(capture_output, buffer=compressed),
            recipients=(key.fingerprint,),
            trust_model=gpg.TrustModel.always,
            compress_algo=gpg.CompressAlgo.ZLIB,
            compress_level=9,
        )

        self.s.encrypt(
            source=message,
            output=partial(capture_output, buffer=not_compressed),
            recipients=(key.fingerprint,),
            trust_model=gpg.TrustModel.always,
            compress_algo=gpg.CompressAlgo.NONE,
        )

        # Check that invalid compression levels raise the expected error.
        for compress_level in (-1, 33, "seven"):
            with self.assertRaises(ValueError):
                self.s.encrypt(
                    source=message,
                    output=partial(capture_output, buffer=compressed),
                    recipients=(key.fingerprint,),
                    trust_model=gpg.TrustModel.always,
                    compress_algo=gpg.CompressAlgo.ZLIB,
                    compress_level=cast(int, compress_level),
                )

        # Check missing passphrase raises an error when signing is requested.
        with self.assertRaises(ValueError):
            self.s.encrypt(
                source=message,
                output=partial(capture_output, buffer=compressed),
                recipients=(key.fingerprint,),
                sign=key.fingerprint,
                passphrase=None,
                trust_model=gpg.TrustModel.always,
                compress_algo=gpg.CompressAlgo.ZLIB,
                compress_level=cast(int, compress_level),
            )

        self.assertLess(len(compressed.read()), len(not_compressed.read()))

    def test_delete_keys(self) -> None:
        self.s.import_file(SGT_HARTMANN_PUB_KEY)
        (key,) = self.s.list_pub_keys()
        self.assertEqual(key.fingerprint, SGT_HARTMANN_FINGERPRINT)
        self.s.delete_pub_keys(key.fingerprint)
        self.assertFalse(self.s.list_pub_keys())

    def test_detach_sig(self) -> None:
        self.s.import_file(CN_SEC_KEY, passphrase="test")
        (cn_key,) = self.s.list_pub_keys()
        (cn_key,) = self.s.list_sec_keys()
        msg = b"Camelot! ... It's only a model."
        with self.s.detach_sig(msg, passphrase="test") as out:
            sig = out.read()
        fpr = self.s.verify_detached_sig(sig=sig, src=msg)
        self.assertEqual(fpr, cn_key.sub_keys[0].fingerprint)
        with self.assertRaises(gpg.GPGError):
            self.s.verify_detached_sig(sig=sig, src=b"wrong message")

    def test_gnupg_home_dir_expansion(self) -> None:
        input_paths = (
            "/some/path",
            "/some/path/",
            "~/some/path/",
            "/home/user/some/~/path",
            "/home/user/some/../path",
            "~/../some/path",
            "",
            None,
        )
        home_dir = os.path.expanduser("~")
        default_gnupg_home_dir = store.get_default_gnupg_home_dir()
        expected_values = (
            "/some/path",
            "/some/path",
            os.path.join(home_dir, "some/path"),
            "/home/user/some/~/path",
            "/home/user/path",
            os.path.join(os.path.dirname(home_dir), "some/path"),
            default_gnupg_home_dir,
            default_gnupg_home_dir,
        )

        # Note: the reason for using os.path.abspath() on the expected_value
        # in the comparison below is to ensure the path uses the correct,
        # plateform specific, path separator.
        for input_path, expected_value in zip(input_paths, expected_values):
            g = store.GPGStore(gnupg_home_dir=input_path)
            self.assertEqual(g.gnupg_home_dir, os.path.abspath(expected_value))


CN_PUB_KEY = b"""-----BEGIN PGP PUBLIC KEY BLOCK-----

mI0EXcPQFwEEAMm1d9jlC5Zw1y/buUV6MuT5ABp4H8Sr3+MUf/GMFKPvd4Qboitd
Yf3eX753y0MNgzFlsSl8CtSO9pd+lqEt4WNMbIVWti04JAM6tD4lqeSzW57vlQlb
sGRf70fg9Qt1dc2LR0HcUi6bw46eWbSGGR3ZM7jbxB9F/VjpaTKI9S69ABEBAAG0
KkNodWNrIE5vcnJpcyA8Y2h1Y2subm9ycmlzQHJvdW5kaG91c2UuZ292PojOBBMB
CAA4FiEEKGahQqFvVOButzz5Lj5/4o/kEq4FAl3D0BcCGy8FCwkIBwIGFQoJCAsC
BBYCAwECHgECF4AACgkQLj5/4o/kEq59GQP+ONPvcLBccHAYbf0utUCiWZux9EwH
BN/TfOo1Vpgbq/DiBZcPoVeDvGTJDPg3dSUxsQzMHztp1UAi6o5Q6XCAGALMlxU9
QIHb4fyfbTQX1QrtB2EyLD/9O8nePrhqTqAd+xpdyvAw8L+hMoTyk0JnVW/UBVLH
RK4L9NaBxhSapfS4jQRdw9AXAQQAsuju5Ccwg3mDJcV+bR9yHIc+XyXcsbzFW0VU
lgNXiQoVHKJQ+X/SgElmxKwik8yeQ4CiM3/GeKmjhKB7JZnp4U+61Dfew+qc6Net
uAdNu7f26x5q237tjU9zx4JzhdnDMhrXkuYl/VR95S7CHlt1QlVxP1/Psx11a/Ej
xB+7ZDsAEQEAAYkBawQYAQgAIBYhBChmoUKhb1Tgbrc8+S4+f+KP5BKuBQJdw9AX
AhsuAL8JEC4+f+KP5BKutCAEGQEIAB0WIQRK0g8wPkscJLLIn7rzDUwzQHo/YwUC
XcPQFwAKCRDzDUwzQHo/YzIkA/4m5TWu7yvVzxjtybqHpShxrrZeTzWRrSu/SiCQ
bR8pGWP9NhnRDcfH/zMX6sMMorE6BBHI8fFMzJB974Lv0q2S8njVMuv+tIH7t+Um
KxlTY2Nz+uzFjuv2rDvPL4ozb4iZfOFq1WOwrml3fpE3qQzqLypd4OBj4H+ITghJ
DiJ55n2eBACItELQ3PcTXeh7XmXTeV3/lVhGHWlhUvcQUCNZCo9AFoX9aYBPX6OF
arsEzjNcZilc/R+r3xYF4kV2kspQQPGoslHN0Zq4Wxp0tOh/IevqjJ3bSzcCqwGz
c9fhdV81kJHiCZUFT03X+D5LrOJJ857/vwL1Roi5UgjRjX1jMFV/tg ==
=0f3v
-----END PGP PUBLIC KEY BLOCK-----
"""

CN_SEC_KEY = b"""-----BEGIN PGP PRIVATE KEY BLOCK-----
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 ==
=G52m
-----END PGP PRIVATE KEY BLOCK-----
"""


UMLAUTS_PUB_KEY = b"""-----BEGIN PGP PUBLIC KEY BLOCK-----

mI0EXcQPZwEEAMFLJh5nhUL3e2cIIEHw2lUwleWu/FMgy5GVVGho1pzSjqxc2cML
0PTLePd5/w/BOHH7tx2404CcwNVF8NrdPgFLV7mZu0Q4eUJNG+yz4EpUBCVK2si/
vNijeF3YH2vO18cyTCGiCW6vXi1cJqf6k73PCrSOPqj7QuNi41Ahkr5FABEBAAG0
KVdpdGggVW1sw6R1dHMgPHdpdGgudW1sYXV0c0B1bWxhdXRzLmluZm8+iM4EEwEI
ADgWIQS1Rlqvu4FLhxmgfbVt8x+lbBrrcQUCXcQPZwIbLwULCQgHAgYVCgkICwIE
FgIDAQIeAQIXgAAKCRBt8x+lbBrrca0uA/9Dj+s3F4FzO2MsDGaD8WXuZ8LPFSRW
zehWewIeRS6hT7L9+LNF3m0GVOueXCPrHj/8iBstc5UbFuph3Pu/3Cgw5YnnSkcC
/LPFGRlc5Z6oRF259qZNW8NafOoi1E4o/LAN1QG5SChYUS2jcyDrosJekGEaPP3n
0KqQQPYC7Smu3riNBF3ED2cBBADClqvryRmApeYO8tQcGWaiFa27vFmM4FCAzbPK
tt88N1GaHWgycgVV/dChhWDVVkEDnas+WwVMz39ZZi8tejBrk1mtoIIBctn8AWJL
W7VH9ZOrfPm1UuqLE7YMhu9hl7i4ANIJnJMsjYnOnb/Z9ButzIobZdigV7a1j47t
pXHnuwARAQABiQFrBBgBCAAgFiEEtUZar7uBS4cZoH21bfMfpWwa63EFAl3ED2cC
Gy4AvwkQbfMfpWwa63G0IAQZAQgAHRYhBHW94ZPbXFu5K4+7zYDWfVUAnJDMBQJd
xA9nAAoJEIDWfVUAnJDMBb4EAKt1sMe/6MhTsn3weshU5U6T217W/lxXmz5vErnO
yd0WaMWOsGN7WEV/UtoVMZ/EBBdIqswUiujDPd8oL3yCsNBZNDigyCGX6shqCeRP
AAUkpC/0KWDP33PNenGSMFzoBcghYBRUuVucvamzskDfidsKbLv+/7FkRwzstg7L
5/GUKGkD/2LlSZlm+qDBzDt5q+tQaGOQnJmebF7o4FQ8feY7GNcFQ9DMV9lR5vuZ
jV4OTQVS2Prv2ASuN7rPRbUYeJD5t7Mko9DzL+6XGgjPim6Wg1xlRSRkskksbX3y
QndrRiimeh43nPBLOBRN9HQoaywkc6hoAZTQAvCgOZli507dp8Mz
=sfjX
-----END PGP PUBLIC KEY BLOCK-----
"""

SGT_HARTMANN_PUB_KEY = b"""-----BEGIN PGP PUBLIC KEY BLOCK-----

mQENBF3MKrABCACw6ZFUve5lTSCtDfhcuDH0pFd7AF0uM1o4kEBLKfcEImrgGo7S
LX7tooxsx2gNv4ARHW4XL5pixLj4itRBDeUGczdDzNwQeE6HmWNW2JtqF7/h1TTr
iw3UwS3eOG8sNrZxx2yLoIF/JZP2CDLz7jXDhuAvLdoK8hzNX82PVUkoVKtPYhbk
1gw3y1coH2Z6rdAp6EwrMSg95vbwzKjuhcyYlxRodwlvmJhDBRxr86QCCRNiA6R8
7Nugtt/wP5XbxcxgNBmciX10NPR4eIoQk4acyGlDvOdrF50h2DPAw3HUdOw513xN
4b7NOHlJJk0+Uo4yKcZ2LXmeiSNU6TY93moLABEBAAG0LUd1bm5lcnkgU2d0LiBI
YXJ0bWFubiA8aGFydG1hbm5AYnVsbHNoaXQuZ292PokBVAQTAQgAPhYhBNoaNjvF
4R2nOAaopOPkAkoOViIeBQJdzCqwAhsDBQkDwmcABQsJCAcCBhUKCQgLAgQWAgMB
Ah4BAheAAAoJEOPkAkoOViIeSIEH/RDXmilCOkoR6uxJgNm7MUqavL2/s3ks57uk
qDrldJm3T5SKBCgrUSP3UaiBIwq+M32k23+CLNl8RQBVwhFAnJcHmMK/fEVcSK8G
HKy4CS8yu1xwDIgWWeDBtbo6r4IEah19dBSLOrxeyBFoI65pRHP18AWMaHnKv5nz
GXdGiLlRl6rKbyBxnesXweR+IIZJhciphhzuV8HtnyHsess0Toq2kjiCGp7Kjkz1
6YQSiHawAj3nV+/RK0kNLZsuQOgY7X2WVw4HAU//Ap1XdtMbOxberLcUU7Rzeudg
uiWfkmTMM36mhojOjtPhgB1J9ZZZCe5KM7hRB2VFuJE45r3S5LCIswQTAQgAHRYh
BChmoUKhb1Tgbrc8+S4+f+KP5BKuBQJdzCsMAAoJEC4+f+KP5BKuz3oD/inKIFnX
KzFblhnLeHf+tW7JQ9EVu+7rTKKyHHay7sgzVQ2KfySemFYwn2fgXUOuZtuk0axe
BHV0qTgL0vfnadxXF96tjFQ14xhBym6KR2ccvQN+/GLM5c0iBbp59azMy0q4VyJH
5c435C3Yq9ownEvJRRdYdJgaFTU1wjwQs7qZuQENBF3MKrABCADguI7FCLGtG6rZ
/EJv58Bnihfz9OuUSc/R6DaFbGGtolTzpTskX3pK1daYigPDDv4iMN5KnO1eeJyp
wmyRpC2yKBLw70z87QGPSviQHWTvyj22YF9ByywFKAuZeY2tpWwkLW2bbVm38LLT
hwmDRkQ8nQZPufvpDOHq6AC84ymW5TuHKEYTfF//14hyonOK40HNjSdEC6ZvfN2n
N80y9g3IvaNGsPOo3MW4z877W9v/IAvFlTEVnyRA0hXEykFnUYk//Eh6cnDTlrCO
hLwd9bbcivPPpjiGYLjgvDghyXK/qv2T4ahiMq2qxp5F8c4x/aTq8U//9GIEU7nd
danxrHsZABEBAAGJATwEGAEIACYWIQTaGjY7xeEdpzgGqKTj5AJKDlYiHgUCXcwq
sAIbDAUJA8JnAAAKCRDj5AJKDlYiHiyUB/9ZH1ZUF/lHsFjFgypyv1BfqFIyYZk9
1/zSHdwUl7svtBrc8XewisxgCBywbzJWHCooAAkSmwSrCFnqVfoUFCdAYxd3lrX1
DA2WEDcIr3Im2wUOKw0fB5yyBcep5QAzoRFl3uZwOcAYrqq38L7xAoRka991oPIk
nzY4GfcR+7ftmzKDPcXjTIbLQwmDNR0vDy7kNBB7ycHWwrMfadWGrGmeQ8pgTnP4
Egfkuv5fEQ3njiV3QcRwmTtqPEVPmA1G/TStAPZa9cBfYHTxH1JzxwVhYvBWXoA3
vwMVdh5kI53+JxM0k7OS6SI8cpf4pmiyLypdkyOCql2HYe7iPuRk0IYi
=lmDh
-----END PGP PUBLIC KEY BLOCK-----
"""

ENCRYPTED_MESSAGE = (
    b"\x84\x8c\x03\xf3\rL3@z?c\x01\x03\xff\\\t\xea;f\xb4`b\x05\xfeJ(\xdf"
    b'^`"\x87B/\x8b\xdfYD\xb4^\x03\x99R\x0bn\x1c\x12\xca6Ju\x8e\x03\x1fI'
    b"U|~\x89>w\x18'\xad\xbb\xa0\x86\x85\x1a\r\x93\xe8$\xd0H\xf8\xef\xfa"
    b"\x16\x87\x17K|\x9d$\xeb<\xbd\x0e3N\xbd{\xf1bT\xcd\xa3-\xc6\x83\x0e"
    b"\x96\xc5\xcb--\x93P\x15\xe7@\x84\x8fR\xfa\t\xc2\x99\x1f\xc4\xf0`\xfc"
    b"\x87v\x04\x1c\x9ef\x03\xc7Jh\x840\x112/\x1b\x06L\xf4\xd2<\x01\xbb"
    b"\xd2\x80=\x17\x9d\xc5ww\n\xd9\xf7%\x9fpzg\\\x1fz\x02\xf8$\x17\xe7V"
    b"\x825\xdd\n\xedu\x12\xa9\xac\xab5\x99\x9b\x7f\x9a\x8f\\\xd1\xa8\xb2"
    b"\xe9!\x1dH\xa3\x1c\xd9\xf9\xd1l\x86\xe7\x12"
)
ENCRYPTED_MESSAGE_CN = (
    b"\x84\x8c\x03\xf3\rL3@z?c\x01\x03\xffw\x058f\x95\xbd\xe5\x8d1q"
    b"\xdd\x80\xbd\x8d\xb5v5\xb3\xcf\x92\xb9?6\x8a \xfc\xc9\xc9\xad"
    b"\xccI\xd7\xc1@\xe3\xb3]\xbag\xcb\xae\x99\xbdkB\xdd\xcaS\x86598"
    b"\x00\xdd\xe1C\xad$1<\xe0\xac\xd5\x9b4\xc0w\xb5O\xf6\x8a/\x88\x7f"
    b"\x1b!\xb7\x08x\x0cUV\xd3\x1e\xaba\xe8\x03]\xc79\xea\x05\x0f*\xdc"
    b"\xa5\x88\xac\xa3Y\xa0\xfb\xd24\xf4\xa0hD\xe3\xaa\x11\xd1\xf9o"
    b"\xf7\xf7\xc3\xc8:\x0c\x16\xd5W\xfa%\x8f7\xd2F\x01q\x07\xd0\xde"
    b"~\xe1V\x00h\xc0O=ls?\x13\xfc\x9e\x8aY>e\x10$\xd2\xdb;\x18\xc0"
    b"\xf7]\xa2)r\xfe\xf8\xd9\x05\xcb\xcc\r8\xb5\xaf\x8b\xa0\xff.\xa1T"
    b"&_\x17\xdb\x0bi\x04\xd0\xa3\xa2\x9c\x8b\x80\x96\xb5\x10I\xea`"
)



import base64
import io
from contextlib import contextmanager
from typing import Optional, Iterator, IO

from .cmd import GPGError
from .model import (
    CompressAlgo,
    Key,
    KeyCapability,
    KeyInfo,
    KeyType,
    RevocationSignature,
    Signature,
    SignatureValidity,
    SubKey,
    TrustModel,
    Uid,
    Validity,
)
from .store import GPGStore, get_default_gnupg_home_dir
from .keyserver import search_keyserver, KeyserverError, KeyserverKeyNotFoundError
from .version import __version__

__all__ = [
    "CompressAlgo",
    "GPGError",
    "GPGStore",
    "Key",
    "KeyCapability",
    "KeyInfo",
    "KeyType",
    "RevocationSignature",
    "Signature",
    "SignatureValidity",
    "SubKey",
    "TrustModel",
    "Uid",
    "Validity",
    "extract_key_id",
    "extract_key_id_from_sig",
    "get_default_gnupg_home_dir",
    "public_key_encrypted_session_key_packet",
    "search_keyserver",
    "signature_packet",
    "KeyserverError",
    "KeyserverKeyNotFoundError",
]


def extract_key_id(src: IO[bytes]) -> Iterator[str]:
    """Extract the public key_id from an encrypted message"""
    try:
        with preserve_pos(src):
            while True:
                fpr = extract_key_id_once(src)
                if fpr is None:
                    break
                yield fpr
    except ValueError as e:
        marker = b"-----BEGIN "
        src.seek(0)
        msg_start = src.read(len(marker))
        if msg_start == marker:
            raise GPGError("ASCII armored encrypted data is not supported") from e
        raise GPGError(f"Corrupted encrypted message: {e}") from e


def public_key_encrypted_session_key_packet(src: IO[bytes]) -> bytes:
    """type 1: Public-Key Encrypted Session Key Packet"""
    _version = src.read(1)
    return src.read(8)


def signature_packet(src: IO[bytes]) -> bytes:
    """Type 2: Signature Packet"""
    (version,) = src.read(1)
    if version == 3:
        (length,) = src.read(1)
        if length != 5:
            raise ValueError(f"Invalid length: {length}. Should be 5")
        _sigtype = src.read(1)
        _creationtime = src.read(4)
        return src.read(8)

    if version == 4:
        # fpr starts at octet 12
        _sigtype = src.read(1)
        _pubkeyalgorithm = src.read(1)
        _hashalgorithm = src.read(1)
        # Version 4 Signature Packet Format contains two blocks of subpackets.
        # Try extracting a fingerprint from the hashed subpackets block.
        # If the fingerprint subpacket is missing, extract a key id from
        # the unhashed subpackets block.
        for _ in range(2):
            octet_count = int.from_bytes(src.read(2), "big")
            subpackets_end = src.tell() + octet_count
            while src.tell() < subpackets_end:
                key_id = read_subpacket(src)
                if key_id:
                    return key_id
        _signedhash = src.read(2)
        raise ValueError(
            "No issuer fingerprint/key_id subpacket found in signature packet"
        )

    raise ValueError(f"Invalid version: {version}. Only 3 and 4 are supported")


def read_subpacket(src: IO[bytes]) -> Optional[bytes]:
    l = read_new_packet_len(src)
    if l is None:
        return None
    (pkg_type,) = src.read(1)
    body = src.read(l - 1)
    if pkg_type == 33:  # type 33: Issuer Fingerprint (rfc4880bis)
        # The first octet is key version number
        return body[1:]
    if pkg_type == 16:  # type 16: Issuer
        return body
    return None


PACKET_HANDLERS = {1: public_key_encrypted_session_key_packet, 2: signature_packet}


def read_new_packet_len(src: IO[bytes]) -> Optional[int]:
    (l,) = src.read(1)
    if l < 192:  # 1 octet length
        return l
    if l <= 223:  # 2 octet length
        (l2,) = src.read(1)
        return ((l - 192) << 8) + l2 + 192
    # 224 <= l < 255: Partial Body Lengths (=> header has length 1)
    if l < 255:
        return None
    if l == 255:  # 5 octet length
        l2, l3, l4, l5 = src.read(4)
        return (l2 << 24) | (l3 << 16) | (l4 << 8) | l5
    raise ValueError(f"Invalid length: {l}")


def extract_key_id_from_sig(src: bytes) -> Optional[str]:
    """Extract key id from a signature packet. Either ascii armored or binary"""
    if src[0] == b"-"[0]:
        src = ascii_armored_to_binary(src)
    try:
        return extract_key_id_once(io.BytesIO(src))
    except ValueError as e:
        raise GPGError(
            "Found an invalid pgp signature while trying to extract a key id"
        ) from e


def extract_key_id_once(src: IO[bytes]) -> Optional[str]:
    """Reads one packet from src.

    According to https://www.ietf.org/rfc/rfc4880.txt and rfc4880bis
    """
    try:
        (packet_hdr,) = src.read(1)
    except ValueError:
        return None
    # 1st bit must be 1
    if not packet_hdr & 0x80:
        raise ValueError("Not a valid packet tag")
    # 2nd bit is packet version (new / old)
    new_version = packet_hdr & 0x40
    packet_type = packet_hdr & 0b00111111
    if not new_version:
        packet_type >>= 2
    pkg_handler = PACKET_HANDLERS.get(packet_type)
    if pkg_handler is None:
        return None
    if not new_version:
        packet_length_type = packet_hdr & 0b11
        body_offs_by_len_type = {0: 2, 1: 3, 2: 5, 3: 1}
        try:
            packet_len_size = body_offs_by_len_type[packet_length_type] - 1
        except KeyError as e:
            raise ValueError(f"Invalid packet length type: {packet_length_type}") from e
        if packet_length_type == 3:  # The packet is of indeterminate length
            packet_len = None
        else:
            packet_len = int.from_bytes(src.read(packet_len_size), "big")
    else:
        packet_len = read_new_packet_len(src)
    if packet_len is not None and packet_len < 10:
        raise ValueError(f"Unexpected packet length of {packet_len}")
    body_start = src.tell()
    pub_key_id = pkg_handler(src)
    if packet_len:
        src.seek(body_start + packet_len)
    else:
        src.seek(0, 2)
    return pub_key_id.hex().upper()


@contextmanager
def preserve_pos(src: IO[bytes]) -> Iterator[IO[bytes]]:
    pos = src.tell()
    yield src
    src.seek(pos)


def ascii_armored_to_binary(src: bytes) -> bytes:
    header, _, *lines, footer = src.splitlines()
    # If the last line is empty, continue to the last not empty line:
    while not footer:
        footer = lines.pop()
    if header != b"-----BEGIN PGP SIGNATURE-----":
        raise ValueError(f"Invalid ascii armored signature header: '{header.decode()}'")
    if footer != b"-----END PGP SIGNATURE-----":
        raise ValueError(f"Invalid ascii armored signature footer: '{footer.decode()}'")
    return base64.decodebytes(b"".join(l for l in lines if l))



import unittest

from gpg_lite import cmd


class TestGPG(unittest.TestCase):
    def test_cmd_pipe(self) -> None:
        with self.assertRaises(FileNotFoundError):
            with cmd.cmd_pipe(("non-existing-bin",)) as proc:
                cmd.assert_io(proc.stdout).read()

        with self.assertRaises((cmd.GPGError, FileNotFoundError)):
            with cmd.cmd_pipe(("python3", "--non-existing-option")) as proc:
                cmd.assert_io(proc.stdout).read()



import re
from enum import Enum, auto
from typing import Optional, AbstractSet, Tuple, List, Union

from dataclasses import dataclass, fields as dataclass_fields, field


@dataclass(frozen=True)
class Uid:
    """GPG User ID representation.

    The GPG user ID format can contain a user's name and/or email address in
    the following formats:
     - Alice (Alice's test PGP key) <alice@example.com>
     - Alice
     - alice@example.com
    """

    full_name: Optional[str]
    email: Optional[str]

    def __post_init__(self) -> None:
        if self.full_name is None and self.email is None:
            raise ValueError("Both 'full_name' and 'email' cannot be None.")

    def __repr__(self) -> str:
        if self.full_name is not None and self.email is not None:
            return f"{self.full_name} <{self.email}>"
        return self.full_name or self.email or "?"

    @staticmethod
    def from_str(s: Optional[str]) -> Optional["Uid"]:
        if s is None or s[0] == "[" and s[-1] == "]":
            # Something like "[User ID not found]" (can be translated into other languages)
            return None
        if re.search(r" <\S+>", s) is not None:
            return Uid(*s.rstrip(">").split(" <"))
        if "@" in s:
            return Uid(full_name=None, email=s)
        return Uid(full_name=s, email=None)


class KeyType(Enum):
    """Representation of the GPG key type. Only two type of keys exist:
    public and secret (often also referred to as 'private')"""

    public = auto()
    secret = auto()


class TrustModel(Enum):
    pgp = "pgp"
    classic = "classic"
    tofu = "tofu"  # Trust on first use.
    tofu_pgp = "tofu+pgp"
    direct = "direct"
    always = "always"
    auto = "auto"


class Validity(Enum):
    unknown_new = "o"  # Unknown (this key is new to the system)
    invalid = "i"  # The key is invalid (e.g. missing self-signature)
    disabled = "d"  # The key has been disabled
    #               # (deprecated - use the 'D' in field 12 instead)
    revoked = "r"  # The key has been revoked
    expired = "e"  # The key has expired
    unknown = "-"  # Unknown validity (i.e. no value assigned)
    undefined = "q"  # Undefined validity.  '-' and 'q' may safely be
    #                # treated as the same value for most purposes
    not_valid = "n"  # The key is not valid
    marginal_valid = "m"  # The key is marginally valid.
    fully_valid = "f"  # The key is fully valid
    ultimately_valid = "u"  # The key is ultimately valid. This often means
    #                       # that the secret key is available, but any key
    #                       # may be marked as ultimately valid.
    well_known_private_part = "w"  # The key has a well known private part.
    special_validity = "s"  # The key has special validity.  This means
    #                       # that it might be self-signed and
    #                       # expected to be used in the STEED system.


class SignatureValidity(Enum):
    good = "!"
    bad = "-"
    no_public_key = "?"
    error = "%"


class CompressAlgo(Enum):
    ZLIB = "zlib"
    ZIP = "zip"
    BZIP2 = "bzip2"
    NONE = "uncompressed"


class KeyCapability(Enum):
    encrypt = "encrypt"
    sign = "sign"
    certify = "cert"
    authentication = "auth"


class SpecialKeyCapability(Enum):
    unknown = "?"


@dataclass(frozen=True)
class Signature:
    """Representation of a signature on a public key.

    When signature's hash algorithm is SHA-1 (and not SHA-256), the signature
    has no fingerprint.
    """

    issuer_uid: Optional[Uid]
    issuer_key_id: str
    issuer_fingerprint: Optional[str]
    signature_class: str
    validity: SignatureValidity
    creation_date: str
    expiration_date: Optional[str] = None


@dataclass(frozen=True)
class RevocationSignature(Signature):
    reason: Optional[str] = None
    comment: Optional[str] = None


@dataclass(frozen=True)
class SubKey:
    key_id: str
    key_type: KeyType
    fingerprint: str
    key_length: int
    pub_key_algorithm: int
    creation_date: str
    validity: Validity = Validity.unknown
    expiration_date: Optional[str] = None
    key_capabilities: AbstractSet[
        Union[KeyCapability, SpecialKeyCapability]
    ] = frozenset()
    signatures: Tuple[Signature, ...] = ()
    disabled: bool = False

    @property
    def valid_signatures(self) -> List[Signature]:
        """Return a list of signatures on a public key that are verified and
        non-revoked.
        """

        revoked_sigs = {
            s.issuer_fingerprint or s.issuer_key_id
            for s in self.signatures
            if isinstance(s, RevocationSignature)
            and s.validity == SignatureValidity.good
        }
        return [
            s
            for s in self.signatures
            if s.issuer_fingerprint not in revoked_sigs
            and s.issuer_key_id not in revoked_sigs
            and s.validity == SignatureValidity.good
        ]


@dataclass(frozen=True)
class Key(SubKey):
    uids: Tuple[Uid, ...] = field(default=(), metadata=dict(required=True))
    owner_trust: str = "-"
    sub_keys: Tuple[SubKey, ...] = ()
    origin: Optional[str] = None
    combined_key_capabilities: AbstractSet[
        Union[KeyCapability, SpecialKeyCapability]
    ] = frozenset()

    def __post_init__(self) -> None:
        """Raise error if there is any required field with no value."""
        missing_fields = [
            f.name
            for f in dataclass_fields(type(self))
            if f.metadata.get("required", False) and not getattr(self, f.name)
        ]
        if missing_fields:
            l = len(missing_fields)
            plural = "s" if l > 1 else ""
            fields = ", ".join(f"'{f}'" for f in missing_fields)
            raise TypeError(f"__init__ missing {l} required argument{plural}: {fields}")


@dataclass(frozen=True)
class KeyInfo:
    uid: Uid
    fingerprint: str
    key_algorithm: int
    key_length: int
    creation_date: str
    expiration_date: Optional[str]



import unittest
from pathlib import Path

from test.test_gpg import ENV_WITH_GPG


@unittest.skipUnless(ENV_WITH_GPG, "Integration tests")
class TestDocs(unittest.TestCase):
    @staticmethod
    def test_gen_key() -> None:
        """Read the python code snippets from README and exec"""
        read = False
        code = ""
        with open(Path(__file__).parent.parent / "README.md", encoding="utf-8") as doc:
            for line in doc:
                if line == "```python\n":
                    read = True
                    continue
                if line == "```\n":
                    read = False
                    continue
                if read:
                    code += line

        exec(code)  # pylint: disable=exec-used



import io
import platform
import os
import re
import tempfile
import shutil
from typing import (
    Optional,
    List,
    Tuple,
    Sequence,
    Iterator,
    Iterable,
    Union,
    Any,
    Type,
    IO,
    Dict,
    ContextManager,
    Callable,
    AbstractSet,
)
import urllib.request

from dataclasses import dataclass, fields as dataclass_fields

from .parser import compile_grammar
from .deserialize import deserialize
from .cmd import (
    cmd_devnull,
    cmd_pipe,
    cmd_pipe_stdout,
    assert_io,
    expect,
    GPGError,
    stderr_lookahead,
    ISource,
    OSource,
)
from .model import (
    Uid,
    KeyType,
    Key,
    SubKey,
    TrustModel,
    SignatureValidity,
    CompressAlgo,
    Signature,
    RevocationSignature,
    Validity,
    KeyCapability,
    SpecialKeyCapability,
)

from .keyserver import upload_keys, download_key, UrlOpener

Token = Union["ColonRecordBase", "Factory", Optional[Uid], str, Signature, list]


unimplemented_col_rec_ids = (
    "crt",  # X.509 certificate
    "crs",  # X.509 certificate and secret key available
    "uat",  # User attribute (same as user id except for field 10).
    "rvs",  # Revocation signature (standalone) [since 2.2.9]
    "pkd",  # Public key data [*]
    "grp",  # Keygrip
    "rvk",  # Revocation key
    "tfs",  # TOFU statistics [*]
    "tru",  # Trust database information [*]
    "spk",  # Signature subpacket [*]
    "cfg",  # Configuration data [*]
)

col_rec_types: Dict[str, Type["ColonRecordBase"]] = {}


def col_rec_type(
    identifier: str,
) -> Callable[[Type["ColonRecordBase"]], Type["ColonRecordBase"]]:
    def decorator(cls: Type["ColonRecordBase"]) -> Type["ColonRecordBase"]:
        col_rec_types[identifier] = cls
        return cls

    return decorator


AnyKeyCapability = Union[KeyCapability, SpecialKeyCapability]


@dataclass
class ColonRecordBase:
    """According to https://github.com/gpg/gnupg/blob/master/doc/DETAILS"""

    validity: Optional[Validity] = None  # 2
    key_length: Optional[int] = None  # 3
    pub_key_algorithm: Optional[int] = None
    key_id: Optional[str] = None  # 5
    creation_date: Optional[str] = None
    expiration_date: Optional[str] = None
    cert_S_N_uid_hash_trust_signature_info: Optional[str] = None
    owner_trust: Optional[str] = None
    user_id: Optional[str] = None  # 10
    signature_class: Optional[str] = None
    key_capabilities: Optional[str] = None
    issuer: Optional[str] = None  # 13: Issuer certificate fingerprint or other info.
    flag: Optional[str] = None
    token_S_N: Optional[str] = None  # 15
    hash_algorithm: Optional[int] = None  # 16: Hash algorithm. 2 = SHA-1, 8 = SHA-256
    curve_name: Optional[str] = None
    compliance_flags: Optional[str] = None
    last_update: Optional[str] = None
    origin: Optional[str] = None  # 20
    comment: Optional[str] = None

    def into(
        self,
    ) -> Token:
        return self


class ColonRecordKey(ColonRecordBase):
    key_type: KeyType = KeyType.public

    def key_fields(self) -> Dict[str, Any]:
        fields = (
            "validity",
            "key_length",
            "pub_key_algorithm",
            "key_id",
            "creation_date",
            "expiration_date",
            "owner_trust",
            "origin",
        )
        field_values = ((f, getattr(self, f, None)) for f in fields)
        return {key: val for key, val in field_values if val is not None}


@col_rec_type("sec")
class ColonRecordSec(ColonRecordKey):
    """Secret key"""

    key_type = KeyType.secret

    def into(self) -> "Factory":
        return Factory(self)


@col_rec_type("pub")
class ColonRecordPub(ColonRecordKey):
    """Public key"""

    def into(self) -> "Factory":
        return Factory(self)


@col_rec_type("sub")
class ColonRecordSub(ColonRecordKey):
    """Subkey (secondary key)"""

    def into(self) -> "SubKeyFactory":
        return SubKeyFactory(self)


@col_rec_type("ssb")
class ColonRecordSsb(ColonRecordKey):
    """Subkey (secondary key)"""

    key_type = KeyType.secret

    def into(self) -> "SubKeyFactory":
        return SubKeyFactory(self)


@col_rec_type("uid")
class ColonRecordUid(ColonRecordBase):
    """User id"""

    def into(self) -> Optional[Uid]:
        return Uid.from_str(self.user_id)


@col_rec_type("fpr")
class ColonRecordFpr(ColonRecordBase):
    """Fingerprint"""

    def into(self) -> str:
        if self.user_id is None:
            raise ValueError("Field 'user_id' missing in record")
        return self.user_id


@col_rec_type("sig")
@dataclass
class ColonRecordSig(ColonRecordBase):
    """Signature"""

    # Here we ignore the change of type in the field. This type is
    # for parsing only.
    validity: Optional[SignatureValidity]  # type: ignore

    def into_dict(self) -> Dict[str, Any]:
        uid = Uid.from_str(self.user_id)
        return dict(
            issuer_uid=uid,
            issuer_key_id=self.key_id,
            issuer_fingerprint=self.issuer,
            creation_date=self.creation_date,
            expiration_date=self.expiration_date,
            signature_class=self.signature_class,
            validity=self.validity,
        )

    def into(self) -> Signature:
        return Signature(**self.into_dict())


@col_rec_type("rev")
@dataclass
class ColonRecordRev(ColonRecordSig):
    """Revocation Signature"""

    def into(self) -> RevocationSignature:
        sig = self.into_dict()
        sig_class = sig.pop("signature_class")
        try:
            sig_class, revocation_reason = sig_class.split(",")
        except ValueError:
            revocation_reason = None
        return RevocationSignature(
            comment=self.comment,
            reason=revocation_reason,
            **sig,
            signature_class=sig_class,
        )


def lex_colon(payload: IO[bytes]) -> Iterator[Token]:
    """Tokenizes the output of a call to the gpg command to list keys (payload)
    one line at a time. Works for both commands to list public or secret keys.
    Example of payload:
        tru::1:1574434353:0:3:1:5
        pub:u:4096:1:12477DE52DAD86CC:1574434343:::u:::escaESCA:
        fpr:::::::::621A2A449F234A0E9246CF8212477DE52DAD86CC:
        uid:u::::1574434343::9F08B014A8964D2FCA0B04A7B11CF5FDFD12DF13::bob <bob@example.com>:
        sig:!::1:34168F55EE0DCAD7:1574434343::::Alice <alice@redqueen.org>:13x::AAA7E2A101EBFA70943DC88534168F55EE0DCAD7:::8:
        sub:u:4096:1:CFAF4D7B40DAF86B:1574434343::::::esa:
        fpr:::::::::39D91E3C8C9C2C0677F1A74BCFAF4D7B40DAF86B:

    :param payload: stream of bytes corresponding to the output of the gpg
        command.
    :return: generator of "token" that are either a "Factory" object or string.
    :raises ValueError:
    """
    # Loop through the output of the GPG command one line at a time.
    for rec in payload:
        rec = rec.rstrip(b"\n")
        if not rec:
            continue

        # Split line by ':' separator. The first element of the line is the
        # "type of record" (e.g. pub = public key, sub = sub key)
        record_type, *args = map(
            lambda s: s if s else None, rec.decode("utf-8", "replace").split(":")
        )
        if record_type is None:
            raise RuntimeError(
                f"Error while parsing row of type {record_type}"
                f"\nGot line without type"
                f"\ncolumns: {args}"
            )
        if record_type in unimplemented_col_rec_ids:
            continue
        kwargs = dict(zip((f.name for f in dataclass_fields(ColonRecordBase)), args))
        try:
            token = deserialize(col_rec_types[record_type])(kwargs).into()
        except BaseException as e:
            raise ValueError(
                f"Error while parsing row of type {record_type}"
                f"\n{format(e)}"
                f"\ncolumns: {args}"
            ) from e
        if token is not None:
            yield token


class Factory:
    type_: Type[SubKey] = Key

    def __init__(self, rec: ColonRecordKey):
        self.kwargs = rec.key_fields()
        self.kwargs.update(self.parse_key_capabilities(rec.key_capabilities))
        self.kwargs["key_type"] = rec.key_type
        if rec.origin:
            origin = rec.origin.replace("\\x3a", ":").lstrip("0123456789 ")
            self.kwargs["origin"] = origin or None

    def append(self, key: str, val: Any) -> "Factory":
        self.kwargs[key] = self.kwargs.get(key, ()) + (val,)
        return self

    def append_subkey(self, val: SubKey) -> "Factory":
        if self.kwargs["key_type"] != val.key_type:
            raise ValueError(f"Wrong sub key type '{val.key_type}'")
        return self.append("sub_keys", val)

    def set(self, key: str, val: Any) -> "Factory":
        if key in self.kwargs:
            raise ValueError(f"Key '{key}' already set")
        self.kwargs[key] = val
        return self

    def build(self) -> SubKey:
        try:
            return self.type_(**self.kwargs)
        except (TypeError, ValueError) as e:
            raise type(e)(e.args[0].replace("__init__", self.type_.__name__))

    @staticmethod
    def parse_key_capabilities(s: Optional[str]) -> Dict[str, Any]:
        mapping: Dict[str, AnyKeyCapability] = {e.value[0]: e for e in KeyCapability}
        mapping.update({e.value[0]: e for e in SpecialKeyCapability})

        if s is None:
            s = ""
        return dict(
            combined_key_capabilities=frozenset(
                mapping[c.lower()] for c in s if c.isupper() and c != "D"
            ),
            key_capabilities=frozenset(mapping[c] for c in s if c.islower()),
            disabled="D" in s,
        )


class SubKeyFactory(Factory):
    type_ = SubKey

    @staticmethod
    def parse_key_capabilities(s: Optional[str]) -> Dict[str, Any]:
        d = Factory.parse_key_capabilities(s)
        del d["combined_key_capabilities"]
        return d


# Factory, type : rule how to transform subsequences of tokens
# into reduced token sequences. For example, in:
#  -> (Factory, str): lambda fac, fpr: [fac.set("fingerprint", fpr)]
# the rule (lambda function) defines how to add the string (fingerprint) to
# the Factory (i.e. the main key object).
key_parser = compile_grammar(
    {
        ((Factory, SubKeyFactory), Uid): lambda fac, uid: [fac.append("uids", uid)],
        ((Factory, SubKeyFactory), str): lambda fac, fpr: [fac.set("fingerprint", fpr)],
        ((Factory, SubKeyFactory), (Signature, RevocationSignature)): lambda fac, sig: [
            fac.append("signatures", sig)
        ],
        (Factory, SubKey): lambda fac, sub: [fac.append_subkey(sub)],
        ((Factory, SubKeyFactory), list): lambda fac, lst: [fac.build(), lst],
        (SubKeyFactory, SubKey): lambda fac, key: [fac.build(), key],
        (Key, list): lambda key, lst: [[key] + lst],
    },
    list,
)


def parse_keys(payload: IO[bytes]) -> List[Key]:
    return key_parser(list(lex_colon(payload)) + [[]])


GPG_DEFAULT_DIR_BY_OS: Dict[str, Tuple[str, ...]] = {
    "Linux": (".gnupg",),
    "Darwin": (".gnupg",),
    "Windows": ("AppData", "Roaming", "gnupg"),
}


def get_default_gnupg_home_dir() -> str:
    """Default path of the directory where GnuPG stores the user's keyrings.
    The location of this directory is platform dependent.
    """
    return os.path.join(
        os.path.expanduser("~"), *GPG_DEFAULT_DIR_BY_OS[platform.system()]
    )


def get_gnupg_binary(gnupg_binary: Optional[str] = None) -> str:
    """Test whether the provided gnupg_binary (name of the GnuPG binary or
    its absolute path) is found in the user's PATH.
    If no input is provided for gnupg_binary, the function defaults to
    testing whether there are binaries named 'gpg' or 'gpg2' in the user's
    PATH.
    """
    binaries_to_test = (gnupg_binary,) if gnupg_binary else ("gpg", "gpg2")
    for bin_name in binaries_to_test:
        try:
            cmd_devnull((bin_name, "--version"))
            return bin_name
        except FileNotFoundError:
            continue

    raise FileNotFoundError(
        f"GnuPG executable [{', '.join(binaries_to_test)}] not found. "
        "Please make sure that the GnuPG binary is found in your PATH. "
        "If your GnuPG binary has a non-standard name (i.e. different from "
        "gpg and gpg2), or is located in a directory outside of your PATH, "
        "please specify it via the 'gnupg_binary' argument of 'GPGStore'."
    )


class GPGStore:
    """Main class providing the functionality of gpg-lite.

    The main class of the library is the representation of a GnuPG 'home'
    directory that contains the user's keyrings.

    :param gnupg_home_dir: 'home' directory of GnuPG, where keyrings are stored
        on the user's machine. On Linux systems, this is typically ~/.gnupg.
        If set to None (default value), the default GnuPG directory is used.
    :param gnupg_binary: name or full path of the GnuPG binary to use. If not
        specified, the user's PATH is searched for a binary named either
        'gpg' or 'gpg2', and an error is raised if none of these in found in
        the PATH.
    """

    def __init__(
        self, gnupg_home_dir: Optional[str] = None, gnupg_binary: Optional[str] = None
    ):
        self._gnupg_home_dir = (
            os.path.abspath(os.path.expanduser(gnupg_home_dir))
            if gnupg_home_dir
            else None
        )
        self._gpg_bin = get_gnupg_binary(gnupg_binary)
        self._version = self.version()
        self._pw_args_t: Tuple[str, ...] = ("--passphrase-fd", "0")
        if self._version >= (2, 2, 0):
            self._pw_args_t = ("--pinentry-mode", "loopback") + self._pw_args_t
        else:
            # Specific to older versions of gpg: the --batch option must be
            # added in order for the --passphrase-fd option to be used.
            self._pw_args_t = ("--batch",) + self._pw_args_t

    def _pw_args(self, pw: Optional[str]) -> Tuple[str, ...]:
        return self._pw_args_t if pw is not None else ()

    @property
    def gnupg_binary(self) -> Optional[str]:
        return shutil.which(self._gpg_bin)

    @property
    def gnupg_home_dir(self) -> str:
        return self._gnupg_home_dir or get_default_gnupg_home_dir()

    @property
    def _base_args(self) -> Tuple[str, ...]:
        args: Tuple[str, ...] = (self._gpg_bin,)
        if self._gnupg_home_dir is not None:
            args += ("--homedir", self._gnupg_home_dir)
        return args

    def default_key(self) -> Optional[str]:
        config_file = os.path.join(self.gnupg_home_dir, "gpg.conf")
        try:
            with open(config_file, encoding="utf-8") as f_cfg:
                default_key = next(
                    line for line in f_cfg if line.startswith("default-key ")
                )
            return default_key.split(" ")[1].rstrip("\r\n")
        except (FileNotFoundError, StopIteration):
            try:
                return self.list_sec_keys()[0].fingerprint
            except IndexError:
                return None

    def list_pub_keys(
        self, search_terms: Optional[Iterable[str]] = None, sigs: bool = False
    ) -> List[Key]:
        """Retrieve public keys from the local keyring that are matching
        one of the search_terms in their key fingerprint, key ID or user ID.

        If no search_terms is passed, all keys are returned.
        Keys are not returned in any particular order.
        If sigs is True, matching keys are returned with their signatures.
        """

        return self._list_keys(
            option="--list-public-keys", search_terms=search_terms, sigs=sigs
        )

    def list_sec_keys(self, search_terms: Optional[Iterable[str]] = None) -> List[Key]:
        """Retrieve secret/private keys from the local keyring that are matching
        one of the search_terms in their key fingerprint, key ID or user ID.

        If no search_terms is passed, all keys are returned.
        Keys are not returned in any particular order.
        If sigs is True, matching keys are returned with their signatures.
        """

        return self._list_keys(option="--list-secret-keys", search_terms=search_terms)

    def _list_keys(
        self,
        option: str,
        search_terms: Optional[Iterable[str]] = None,
        sigs: bool = False,
    ) -> List[Key]:
        """Returns keys from the local keyring that match one of the
        search_terms. The 'options' argument allows to search for either public
        or secret/private keys.
        Important: keys are not returned in the same order as the search_terms.

        :param option: argument to pass to gpg, e.g. "--list-public-keys" to
            list the public keys in the user's keyring.
        :param search_terms: key name, key ID or fingerprint to search for. If
            an empty list is passed, no key is returned. If keys is None, all
            keys in the keyring are returned.
        :param sigs: if True, show signatures appended on keys.
        :return: list of keys matching the search criteria.
        :raise GPGError:
        """
        if search_terms is None:
            search_terms = ()
        else:
            search_terms = tuple(search_terms)
            if not search_terms:
                return []
        # Add arguments that are always required for gpg to produce the
        # correct output. Note that for gpg versions < 2.1, the
        # --with-fingerprint option is required to be passed twice. If passed
        # only once the fingerprint of the subkeys are not printed.
        args: Tuple[str, ...] = (
            option,
            "--fingerprint",
            "--fingerprint",
            "--with-colons",
        )
        if sigs:
            args += ("--with-sig-check",)
        args += search_terms
        try:
            with cmd_pipe(self._base_args + args) as proc:
                return parse_keys(assert_io(proc.stdout))
        except GPGError as e:
            if "error reading key" in str(e).lower():
                return []
            raise

    def encrypt(
        self,
        source: Optional[ISource],
        recipients: Sequence[str],
        output: Optional[OSource],
        sign: Optional[str] = None,
        passphrase: Optional[str] = None,
        trust_model: TrustModel = TrustModel.pgp,
        compress_algo: Optional[CompressAlgo] = None,
        compress_level: Optional[int] = None,
    ) -> None:
        """Encrypt data with the public PGP key of recipients.

        Both source and output can be file like objects or callables
        processing streams.

        :param source: data to encrypt. Read directly from a file object
            or write the data to the stream in a callable.
        :param recipients: gpg key fingerprints.
        :param output: encrypted data. Write directly to a file object
            or pass the stream to a callable.
        :param sign: fingerprint of private key with which the data should be
            signed. If None, the encrypted file is not signed.
        :param passphrase: passphrase of the private key to be used to sign
            the data. Only needed if a value was passed to the sign argument.
            If the private key has no passphrase (bad practice), an empty
            string should be passed as passphrase.
        :param trust_model: trust model to be followed by gpg.
        :param compress_algo: algorithm to use for data compression.
            If not provided the default algorithm defined by gpg is used.
        :param compress_level: data compression algorithm level.
            An integer value between 0 (no compression) and
            9 (maximum compression). If not provided the default level
            defined by gpg is used.
        :raises ValueError:
        """

        # Generate list of recipients for the encrypted file.
        args: Tuple[str, ...] = sum((("--recipient", rcp) for rcp in recipients), ())
        args += ("--no-tty", "--trust-model", trust_model.name, "--encrypt")
        if sign is not None:
            if passphrase is None:
                raise ValueError("Passphrase for private key is missing")
            args += ("--sign", "--local-user", sign)

        if compress_algo is not None:
            args += ("--compress-algo", compress_algo.value)

        if compress_level is not None:
            if (
                not isinstance(compress_level, int)
                or compress_level < 0
                or compress_level > 9
            ):
                raise ValueError(
                    "Invalid compression level. The value must "
                    "be an integer between 0 and 9."
                )
            args += ("--compress-level", str(compress_level))

        maybe_stdout = {} if output is None else {"stdout": output}
        with cmd_pipe(
            self._base_args + self._pw_args(passphrase) + args,
            src=source,
            **maybe_stdout,
            passphrase=passphrase,
        ):
            pass

    def decrypt(
        self,
        source: ISource,
        output: Optional[OSource],
        passphrase: Optional[str] = None,
        trust_model: TrustModel = TrustModel.pgp,
    ) -> List[str]:
        """Decrypt data and, if present, retrieve the signature appended to
        the encrypted file.

        :param source: data to decrypt. Read directly from a file object
            or write the data to the stream in a callable.
        :param output: decrypted data. Write directly to a file object
            or pass the stream to a callable.
        :param passphrase: passphrase of the private key to be used to decrypt
            the data.
        :param trust_model: trust model to be followed by gpg.
        :return: fingerprint or keyid of the signee's key.
        """

        # Note: the '--keyid-format', 'long' argument is only needed for
        # backwards compatibility with gpg versions < 2.1.0
        args = (
            "--batch",
            "--no-tty",
            "--status-fd",
            "2",
            "--trust-model",
            trust_model.name,
            "--decrypt",
        )
        maybe_stdout = {} if output is None else {"stdout": output}
        with cmd_pipe(
            self._base_args + self._pw_args(passphrase) + args,
            **maybe_stdout,
            src=source,
            passphrase=passphrase,
        ) as proc:
            while True:
                if proc.poll() is not None:
                    break
            gpg_out = stderr_lookahead(proc)
            assert_mdc(gpg_out)
            return capture_fpr(gpg_out)

    def gen_key(
        self,
        full_name: str,
        email: str,
        passphrase: Optional[str],
        key_type: str = "RSA",
        key_length: int = 4096,
        key_capabilities: AbstractSet[KeyCapability] = frozenset(
            (KeyCapability.sign, KeyCapability.certify)
        ),
        key_curve: Optional[str] = None,
        subkey_type: str = "RSA",
        subkey_length: int = 4096,
        subkey_capabilities: AbstractSet[KeyCapability] = frozenset(
            (KeyCapability.encrypt,),
        ),
        subkey_curve: Optional[str] = None,
    ) -> str:
        """Create a new public/private PGP key pair.

        :param full_name: user name to be used in the key user ID.
        :param email: email to be used in the key user ID.
        :param passphrase: passphrase (password) that will be requested when
            using the private key. Explicitely set to None if you want an
            unprotected key.
        :param key_length: bit-length of the key to be generated. Generally
            speaking, longer keys are stronger (more difficult to break). For
            "RSA" keys the recommended length is currently 2048 or 4096. Keys
            shorter than 2048 bits should not be used any longer.
        :param key_type: "RSA" (default), "DSA" or other supported gpg algorithms
            (e.g. eddsa, ecdh for ECC keys).
            It is strongly recommended
            to use "RSA", as these key types are more secure.
        :param key_capabilities: The available capabilities are defined in the enum :KeyCapability:
        :param subkey_curve: The requested elliptic curve of the generated key. This is a required parameter for ECC keys. It is ignored for non-ECC keys.
        :param subkey_type: Type for the subkey (See description for :key_type:).
        :param subkey_length: Length for subkey (See description for :key_length:)
        :param subkey_capabilities: See description for :key_capabilities:)
        :param subkey_curve: Key curve for a subkey; similar to :key_curve:.
        :return: fingerprint of the newly generated key.

        Example: To create a ECC key, use:
          key_type="eddsa",
          subkey_type="ecdh",
          key_curve="Ed25519",
          subkey_curve="Curve25519"
        """

        with cmd_pipe(
            self._base_args + ("--batch", "--gen-key", "--status-fd", "1"),
            src=(
                f"""
Key-Type: {key_type}
Key-Length: {key_length}
Subkey-Type: {subkey_type}
Subkey-Length: {subkey_length}
Name-Real: {full_name}
Name-Email: {email}
Expire-Date: 0"""
                + (
                    f"""
Key-Usage: {",".join(c.value for c in key_capabilities)}
Subkey-Usage: {",".join(c.value for c in subkey_capabilities)}"""
                    if self._version >= (2, 2, 0)
                    else ""
                )
                + f"""
{"Key-Curve: " + key_curve if key_curve is not None else ""}
{"Subkey-Curve: " + subkey_curve if subkey_curve is not None else ""}
{"Passphrase: " + passphrase if passphrase is not None else "%no-protection"}
%commit
"""
            ).encode(),
        ) as proc:
            stdout_lines = (
                assert_io(proc.stdout).read().decode("utf-8", "replace").splitlines()
            )
            try:
                l: str = next(
                    line
                    for line in stdout_lines
                    if line.startswith("[GNUPG:] KEY_CREATED")
                )
            except StopIteration:
                raise GPGError(stdout_lines) from None
            return l.split(" ")[-1]

    def send_keys(
        self,
        *fingerprints: str,
        keyserver: str,
        url_opener: UrlOpener = urllib.request.urlopen,
    ) -> None:
        """Upload keys from the user's local keyring to a keyserver."""
        upload_keys(
            keys_as_ascii_armor=self.export(*fingerprints),
            keyserver=keyserver,
            url_opener=url_opener,
        )

    def recv_keys(self, *fingerprints: str, keyserver: str, **kwargs: Any) -> None:
        """Download keys from a keyserver to the user's local keyring."""
        for fingerprint in fingerprints:
            with download_key(
                keyserver=keyserver, fingerprint=fingerprint, **kwargs
            ) as key:
                self.import_file(key.read())

    def delete_pub_keys(self, *fingerprints: str) -> None:
        """Delete one or more public keys with the specified fingerprints from
        the local keyring.

        Note that public keys can only be deleted if there is not matching
        secret key.
        """
        self._delete_keys(*fingerprints, public_key=True)

    def delete_sec_keys(self, *fingerprints: str) -> None:
        """Delete one or more secrte/private keys with the specified
        fingerprints from the local keyring.
        WARNING: deleted secret keys cannot be recovered or re-generated. Tread
        carefully.
        """
        self._delete_keys(*fingerprints, public_key=False)

    def _delete_keys(self, *fingerprints: str, public_key: bool = True) -> None:
        cmd_devnull(
            self._base_args
            + (
                "--batch",
                "--yes",
                "--delete-keys" if public_key else "--delete-secret-keys",
            )
            + fingerprints
        )

    def import_file(self, source: ISource, *, passphrase: Optional[str] = None) -> None:
        cmd_devnull(
            self._base_args + self._pw_args(passphrase) + ("--import", "--batch", "-"),
            src=source,
            passphrase=passphrase,
        )

    def export(self, *fingerprints: str, armor: bool = True) -> bytes:
        cmd = (
            self._base_args
            + ("--export",)
            + (("--armor",) if armor else ())
            + fingerprints
        )
        with cmd_pipe_stdout(cmd) as stream:
            return stream.read()

    def version(self) -> Tuple[int, ...]:
        with cmd_pipe(self._base_args + ("--version",)) as proc:
            return tuple(
                map(
                    int,
                    assert_io(proc.stdout)
                    .readline()
                    .rstrip(b"\n")
                    .split(b" ")[-1]
                    .split(b"."),
                )
            )

    def gen_revoke(self, fingerprint: str, passphrase: str) -> bytes:
        """Generate a revocation certificate for a PGP key with the specified
        fingerprint.
        """
        args = (
            self._base_args
            + self._pw_args(passphrase)
            + (
                "--no-tty",
                "--status-fd",
                "2",
                "--command-fd",
                "0",
                "--gen-revoke",
                fingerprint,
            )
        )
        with expect(args) as proc:
            try:
                proc.put(passphrase.encode() + b"\n")
                proc.expect(b"[GNUPG:] GET_BOOL gen_revoke.okay\n")
                proc.put(b"y\n")
                proc.expect(b"[GNUPG:] GET_LINE ask_revocation_reason.code\n")
                proc.put(b"0\n")
                proc.expect(b"[GNUPG:] GET_LINE ask_revocation_reason.text\n")
                proc.put(b"\n")
                proc.expect(b"[GNUPG:] GET_BOOL ask_revocation_reason.okay\n")
                proc.put(b"y\n")
            except ValueError as e:
                raise GPGError("Failed to generate a revocation certificate") from e
            return assert_io(proc.stdout).read()

    def revoke_key(
        self,
        fingerprint: str,
        passphrase: str,
        revocation_certif: Optional[ISource] = None,
        keyserver: Optional[str] = None,
    ) -> None:
        """Revoke a PGP key, optionally pushing it to a keyserver.

        Key revocation can be done by using a pre-existing revocation
        certificate (best practice is to generate such a certificate when a
        new key is generated, so it can be revoked if the password or secret
        key is lost).
        Alternatively, a key can also be revoked by passing the password of the
        key to revoke. Note that this only works if the matching secret key is
        present in the user's keyring.

        :param fingerprint: fingerprint of the key to revoke.
        :param revocation_certif: revocation certificate associated to the key.
        :param passphrase: password associated to the key to revoke. The
            password is only needed if no value is passed for revocation_certif.
        :param keyserver: URL of keyserver. This argument is optional. If
            provided, then the revoked key is pushed to the specified keyserver.
        """
        if not revocation_certif:
            revocation_certif = self.gen_revoke(fingerprint, passphrase)
        self.import_file(revocation_certif)
        if keyserver is not None:
            self.send_keys(fingerprint, keyserver=keyserver)

    def detach_sig(
        self,
        src: ISource,
        passphrase: str,
        signee: Optional[str] = None,
    ) -> ContextManager[IO[bytes]]:
        """Generate a detached signature for a given input.

        :param src: binary data for which a signature should be generated.
        :param passphrase: passphrase of the private key that is being used to
            sign the key.
        :param signee: fingerprint, key ID or user ID of the PGP key to be used
            to sign the data. If None, the default PGP key in the user's local
            keyring is used.
        :return: a context manager for a buffered reader.
        """
        return cmd_pipe_stdout(
            self._base_args
            + self._pw_args(passphrase)
            + (("--local-user", signee) if signee is not None else ())
            + ("--detach-sig", "-"),
            src=src,
            passphrase=passphrase,
        )

    def verify_detached_sig(
        self, src: Optional[Union[bytes, io.FileIO]], sig: bytes
    ) -> str:
        """Verify and retrieve the PGP signature made to a file in detached
        signature mode - i.e. the signed file and the signature are in separate
        files.

        :param src: file that was signed.
        :param sig: detached signature file.
        :return: fingerprint or key ID of the signee's key.
        :raises GPGError:
        """
        with tempfile.NamedTemporaryFile(delete=False) as f:
            try:
                f.write(sig)
                f.close()
                # Note: the '--keyid-format', 'long' argument is only needed for
                # backwards compatibility with gpg versions < 2.1.0
                with cmd_pipe(
                    self._base_args + ("--status-fd", "2", "--verify", f.name, "-"),
                    src=src,
                ) as proc:
                    gpg_out = stderr_lookahead(proc)
                    fingerprints = capture_fpr(gpg_out)

                    try:
                        (fingerprint,) = fingerprints
                    except ValueError as e:
                        raise GPGError(
                            f"{'More than one' if fingerprints else 'No'} "
                            "fingerprint found in input data:\n"
                            + gpg_out.decode("utf-8", "replace")
                        ) from e
                    return fingerprint
            finally:
                os.remove(f.name)


def capture_fpr(b: bytes) -> List[str]:
    """Extract fingerprint from a standard output of gpg that looks like:
    [GNUPG:] NEWSIG
    gpg: Signature made Tue 10 Mar 2020 11:29:27 CET
    gpg:                using RSA key 3B3EB1983FAA63C1B05E072D066E042415D8FD9C
    [GNUPG:] KEY_CONSIDERED 3B3EB1983FAA63C1B05E072D066E042415D8FD9C 0
    [GNUPG:] SIG_ID PaAuiNMUyUDRMsctXpV+7rX+0CI 2020-03-10 1583836167
    [GNUPG:] KEY_CONSIDERED 3B3EB1983FAA63C1B05E072D066E042415D8FD9C 0
    [GNUPG:] GOODSIG 066E042415D8FD9C Chuck Norris <chucknorris@roundhouse.gov>
    gpg: Good signature from "Chuck Norris <chucknorris@roundhouse.gov>" [unknown]
    [GNUPG:] VALIDSIG 3B3EB1983FAA63C1B05E072D066E042415D8FD9C 2020-03-10 1583836167

    :param b: gpg stdout in binary form.
    :return: either a 40-char long fingerprint.
    """
    return [
        m.group(1).decode("utf-8", "replace")
        for m in re.finditer(rb"\[GNUPG:\] VALIDSIG ([0-9A-F]{40})", b)
    ]


def assert_mdc(b: bytes) -> None:
    """Assert that a valid MDC is present in the GPG output.

    Modification Detection Code is used to detect if the encrypted message
    has not been modified.

    :param b: gpg stdout in binary form.
    :raises GPGError: raised if GOODMDC is absend in the input.
    """
    if b"[GNUPG:] GOODMDC" not in b:
        raise GPGError("Message integrity could not be verified.")
